from datetime import date
import os

from anchorpoint import TextQuoteSelector, TextPositionSelector
from anchorpoint.textselectors import TextPositionSet, TextSelectionError, Range
from dotenv import load_dotenv
from pydantic import ValidationError
import pytest

from legislice.citations import CodeLevel
from legislice.download import Client
from legislice.enactments import (
    CitingProvisionLocation,
    Enactment,
    EnactmentPassage,
    InboundReference,
    TextVersion,
    consolidate_enactments,
)

load_dotenv()

TOKEN = os.getenv("LEGISLICE_API_TOKEN")
API_ROOT = "https://authorityspoke.com/api/v1"


class TestMakeEnactment:
    def test_init_enactment_without_nesting(self):
        s1 = Enactment(
            node="/test/acts/47/1",
            start_date=date(1935, 4, 1),
            heading="Short title",
            content=(
                "This Act may be cited as the Australian Beard Tax"
                " (Promotion of Enlightenment Values) Act 1934."
            ),
        )
        assert s1.end_date is None
        assert s1.nested_children == []
        assert "Enactment(node='/test/acts/47/1'" in repr(s1)

    def test_init_enactment_with_nesting(self):
        subsection = Enactment(
            heading="",
            text_version="The beardcoin shall be a cryptocurrency token…",
            node="/test/acts/47/6C/1",
            start_date=date(2013, 7, 18),
        )

        section = Enactment(
            heading="Issuance of beardcoin",
            text_version="Where an exemption is granted under section 6…",
            node="/test/acts/47/6C",
            children=[subsection],
            end_date=None,
            start_date=date(1935, 4, 1),
        )

        assert section.children[0].content.startswith("The beardcoin shall")

    def test_create_TextPositionSet_on_init(self, section_11_subdivided):
        len_s11 = len(section_11_subdivided["text_version"]["content"])
        data = {
            "enactment": section_11_subdivided,
            "selection": {"positions": [{"start": 0, "end": len_s11}]},
        }
        start_point = len_s11 + len(
            section_11_subdivided["children"][0]["text_version"]["content"]
        )

        margin = 2
        expected = "hairdressers"
        data["selection"]["positions"].append(
            {"start": start_point + margin, "end": start_point + margin + len(expected)}
        )
        result = EnactmentPassage(**data)
        assert isinstance(result.selection, TextPositionSet)
        assert (
            result.selected_text()
            == "The Department of Beards may issue licenses to such…hairdressers…"
        )


class TestLinkedEnactment:
    @pytest.mark.vcr
    def test_text_sequence_for_linked_enactment(self, test_client):
        enactment = test_client.read(query="/test", date="2020-01-01")
        assert "for documentation." in enactment.text
        selected = enactment.select("for documentation.")
        assert selected.selected_text() == "…for documentation."
        assert "for documentation" in selected.text

    def test_enactment_without_children(self):
        enactment = Enactment(
            node="/test/golden",
            heading="The Golden Rule",
            text_version="Do unto others as you would have them do to you.",
            start_date=date(1, 1, 1),
        )
        assert enactment.children == []
        new = enactment.select("Do unto others")
        assert new.selected_text() == "Do unto others…"

    def test_error_blank_content(self):
        with pytest.raises(ValueError):
            Enactment(
                node="/test/unwritten",
                heading="The Unwritten Rule",
                textversion=TextVersion(content=""),
                start_date=date(2001, 1, 1),
            )

    @pytest.mark.vcr
    def test_csl_cite_for_usc(self, test_client):
        enactment = test_client.read(query="/us/usc", date="2020-01-01")
        assert '"container-title": "U.S. Code"' in enactment.csl_json()


class TestEnactmentDetails:
    @pytest.mark.vcr
    def test_usc_enactment_is_statute(self, test_client):
        enactment = test_client.read(query="/us/usc/t17/s103", date="2020-01-01")
        assert enactment.sovereign == "us"
        assert enactment.level == CodeLevel.STATUTE

    @pytest.mark.vcr
    def test_str_representation(self, fourth_a, test_client):
        enactment = test_client.read_from_json(fourth_a)
        selection = TextQuoteSelector(
            exact="The right of the people to be secure in their persons"
        )
        new = enactment.select(selection)
        assert new.level == CodeLevel.CONSTITUTION
        assert new.start_date() == date(1791, 12, 15)
        assert "secure in their persons…" in str(new)
        assert enactment.node in str(new)
        assert "1791-12-15" in str(new)

    def test_node_parts(self, section_11_subdivided):
        enactment = Enactment(**section_11_subdivided)
        passage = enactment.select_all()
        assert passage.section == "11"
        assert passage.title == "47"
        assert passage.code == "acts"
        assert passage.jurisdiction == "test"
        assert passage.sovereign == "test"

    def test_csl_json_fields(self, test_client, section_11_subdivided):
        section = test_client.read_from_json(section_11_subdivided)
        cite_json = section.csl_json()
        assert "event-date" in cite_json

    @pytest.mark.vcr
    def test_sovereign_representation(self, test_client):
        enactment = test_client.read(query="/us")
        assert enactment.code is None
        assert enactment.jurisdiction == "us"

    @pytest.mark.vcr
    def test_constitution_effective_date(self, test_client):
        ex_post_facto_provision = test_client.read(query="/us/const/article/I/8/8")
        assert ex_post_facto_provision.start_date == date(1788, 9, 13)

    @pytest.mark.vcr
    def test_date_and_text_from_path_and_regime(self, test_client):
        """
        This tests different parsing code because the date is
        in the format "dated the 25th of September, 1804"

        This also verifies that the full text of the section
        is selected as the text of the Enactment, even though no
        ``exact``, ``prefix``, or ``suffix`` parameter was
        passed to the TextQuoteSelector constructor.
        """
        amendment_5 = test_client.read(query="/us/const/amendment/V")
        assert amendment_5.start_date == date(1791, 12, 15)
        assert "otherwise infamous crime" in amendment_5.text

    def test_compare_effective_dates(self, fifth_a, fourteenth_dp, test_client):
        amendment_5 = test_client.read_from_json(fifth_a)
        amendment_14 = test_client.read_from_json(fourteenth_dp)
        assert amendment_14.start_date == date(1868, 7, 28)
        assert amendment_5.start_date < amendment_14.start_date

    @pytest.mark.skip(reason="No regulations available via API.")
    def test_regulation_level(self, test_client):
        enactment = test_client.read("/us/cfr/t37/s202.1")
        assert enactment.level == "regulation"

    def test_compare_inbound_citations_on_same_date(self):

        inbound_ref = InboundReference(
            content="or as provided in Section 5",
            reference_text="Section 5",
            target_uri="/test/acts/47/5",
            locations=[
                CitingProvisionLocation(
                    node="/test/acts/47/3/1", start_date=date(2000, 2, 2)
                ),
                CitingProvisionLocation(
                    node="/test/acts/47/7", start_date=date(2000, 2, 2)
                ),
            ],
        )
        assert inbound_ref.latest_location().node == "/test/acts/47/7"


class TestCrossReferences:
    @pytest.mark.vcr()
    def test_no_local_cross_references(self, test_client):
        enactment = test_client.read("/test/acts/47/6D")
        citations = enactment.citations
        assert len(citations) == 0

    @pytest.mark.vcr()
    def test_collect_nested_cross_references(self, test_client):
        enactment = test_client.read("/test/acts/47/6D")
        citations = enactment.cross_references()
        assert len(citations) == 1
        assert citations[0].target_uri == "/test/acts/47/6C"
        assert (
            str(citations[0])
            == 'CrossReference(target_uri="/test/acts/47/6C", reference_text="Section 6C")'
        )

    @pytest.mark.vcr()
    def test_locations_of_cross_reference(self, test_client):
        enactment = test_client.read("/test/acts/47/6D")
        assert "Section 6C" in enactment.text
        references = enactment.cross_references()
        citations = test_client.citations_to(references[0])
        assert citations[0].target_uri == "/test/acts/47/6C"


class TestSelectText:
    client = Client(api_token=TOKEN)

    def test_same_quotation_from_enactments_of_differing_depths(
        self, test_client, section_11_subdivided
    ):
        section = test_client.read_from_json(section_11_subdivided)
        selection = section.select(
            TextQuoteSelector(exact="as they see fit to purchase a beardcoin")
        )
        subsection = test_client.read_from_json(section_11_subdivided["children"][3])
        sub_selection = subsection.select(
            TextQuoteSelector(exact="as they see fit to purchase a beardcoin")
        )
        assert sub_selection.means(selection)
        assert selection >= sub_selection
        assert not selection > sub_selection
        assert not selection.text_sequence() > sub_selection.text_sequence()

    def test_select_text_of_parent_section(self):
        subsection = Enactment(
            heading="",
            text_version="The beardcoin shall be a cryptocurrency token…",
            node="/test/acts/47/6C/1",
            start_date=date(2013, 7, 18),
        )

        section = Enactment(
            heading="Issuance of beardcoin",
            text_version="Where an exemption is granted",
            node="/test/acts/47/6C",
            children=[subsection],
            end_date=None,
            start_date=date(1935, 4, 1),
        )
        selection = section.select(section.content)
        assert selection.selected_text() == "Where an exemption is granted…"
        assert "cryptocurrency" not in selection.selected_text()
        assert str(section) == "/test/acts/47/6C (1935-04-01)"
        assert str(subsection) == "/test/acts/47/6C/1 (2013-07-18)"

    def test_text_sequence_selected_with_bool(self):
        section = Enactment(
            heading="Issuance of beardcoin",
            text_version="Where an exemption is granted…",
            node="/test/acts/47/6C",
            children=[],
            end_date=None,
            start_date=date(1935, 4, 1),
        )
        selection = section.select(section.text_version.content)
        assert selection.text_sequence()[0].text == "Where an exemption is granted…"

    def test_select_with_list_of_strings(self, test_client, section_8):
        section = test_client.read_from_json(section_8)
        selection = section.select(
            [
                "Where an officer of the",
                "state or territorial police",
                "finds a person to be wearing a beard",
                "that officer shall in the first instance issue such person a notice to remedy.",
            ]
        )
        selected_text = selection.selected_text()
        assert selected_text.startswith(
            "Where an officer of the…state or territorial police…finds"
        )

    def test_str_for_text_sequence(self, test_client, section_11_subdivided):
        section = test_client.read_from_json(section_11_subdivided)
        quotes = [
            TextQuoteSelector(
                exact="The Department of Beards may issue licenses to such"
            ),
            TextQuoteSelector(exact="hairdressers", suffix=", or other male grooming"),
            TextQuoteSelector(exact="as they see fit"),
        ]
        selection = section.select(quotes)
        text_sequence = selection.text_sequence()
        assert str(text_sequence) == (
            "The Department of Beards may issue "
            "licenses to such…hairdressers…as they see fit…"
        )

    def test_no_double_spaces_around_repealed_section(self, test_client, section_8):
        section = test_client.read_from_json(section_8["children"][1])
        assert "or remove the beard with" in section.text
        assert "or  remove the beard with" not in section.text

    def test_select_space_between_selected_passages(self, fourteenth_dp, test_client):
        """Test that the space between "property," and "without" is selected."""
        section_1 = test_client.read_from_json(fourteenth_dp)
        selection = section_1.select("without due process of law")
        selection.select_more("life, liberty, or property,")
        now_selected = selection.selected_text()
        assert "or property, without" in now_selected


class TestSelectFromEnactment:
    client = Client(api_token=TOKEN)

    def test_text_of_enactment_subset(self, section_11_together):
        combined = Enactment(**section_11_together)
        selector = TextQuoteSelector(
            exact="barbers, hairdressers, or other male grooming professionals"
        )
        passage = combined.select(selector)
        sequence = passage.text_sequence()
        assert str(sequence).strip("…").startswith("barbers")

    def test_get_passage(self, section_11_subdivided):
        """
        Use selector to get passage from Enactment without changing which part is selected.

        Checks that `.selected_text()` is the same before and after `.get_passage()`.
        """
        section = Enactment(**section_11_subdivided)
        passage = section.select(TextPositionSelector(start=61, end=73))
        assert passage.selected_text() == "…hairdressers…"
        fit_passage = section.get_string(TextPositionSelector(start=112, end=127))
        assert fit_passage == "…as they see fit…"

    def test_get_child_passage(self, section_11_subdivided):
        """
        Use selector to get passage from child of Enactment.
        """
        section = Enactment(**section_11_subdivided)
        passage = section.select(
            [
                TextQuoteSelector(
                    exact="The Department of Beards may issue licenses to"
                ),
                TextQuoteSelector(exact="professionals"),
            ]
        )
        assert passage.node == "/test/acts/47/11"
        assert passage.selected_text().endswith("licenses to…professionals…")
        assert len(passage.selection.ranges()) == 2
        assert passage.selection.ranges()[0] == Range(start=0, end=46)
        assert passage.selection.ranges()[1] == Range(start=98, end=111)

        child_passage = passage.child_passages[2]
        assert len(child_passage.selection.ranges()) == 1
        assert child_passage.selected_text() == "…professionals"
        assert child_passage.selection.ranges()[0] == Range(start=20, end=33)
        assert child_passage.node == "/test/acts/47/11/iii"

    def test_select_nested_text_with_positions(self, section_11_subdivided):
        phrases = TextPositionSet(
            positions=[
                TextPositionSelector(start=0, end=51),
                TextPositionSelector(start=61, end=73),
                TextPositionSelector(start=112, end=127),
            ],
        )
        section = Enactment(**section_11_subdivided)
        passage = section.select(phrases)
        text_sequence = passage.text_sequence()
        assert str(text_sequence) == (
            "The Department of Beards may issue licenses to "
            "such…hairdressers…as they see fit…"
        )

    def test_selection_in_middle_of_enactment(self, test_client, fourth_a):
        result = test_client.read_from_json(fourth_a)
        selector = TextQuoteSelector(
            prefix="and", exact="the persons or things", suffix="to be seized."
        )
        passage = result.select(selector)
        assert passage.selected_text().endswith("or things…")

    def test_select_all_of_empty_enactment(self):
        empty = Enactment(
            node="/us/usc/t1/s101",
            heading="No text",
            text_version="",
            start_date=date(1935, 4, 1),
        )
        passage = empty.select_all()
        assert passage.selected_text() == ""

    def test_select_none(self, section_11_subdivided):
        combined = Enactment(**section_11_subdivided)
        passage = combined.select(False)
        assert passage.selected_text() == ""

    def test_select_none_with_None(self, section_11_subdivided):
        combined = Enactment(**section_11_subdivided)
        passage = combined.select(None)
        assert passage.selected_text() == ""

    def test_select_all(self, section_11_subdivided):
        """Clear selected text, and then select one subsection."""
        combined = Enactment(**section_11_subdivided)
        passage = combined.select(None)
        assert passage.enactment.node == "/test/acts/47/11"
        sub_passage = passage.enactment.children[3].select_all()
        assert (
            sub_passage.selected_text()
            == "as they see fit to purchase a beardcoin from a customer"
        )
        assert sub_passage.enactment.node == "/test/acts/47/11/iii-con"

    def test_select_all_nested(self, section_11_subdivided):
        """Clear selected text, and then select one subsection."""
        section = Enactment(**section_11_subdivided)
        passage = section.select()
        assert passage.selected_text().startswith("The Department of Beards")

    def test_error_for_unusable_selector(self, section_11_subdivided):
        section = Enactment(**section_11_subdivided)
        selection = TextPositionSet(
            positions=[
                TextPositionSelector(start=0, end=10),
                TextPositionSelector(start=1000, end=1010),
            ]
        )
        with pytest.raises(ValueError):
            section.children[3].select(selection)

    @pytest.mark.vcr
    def test_get_positions_from_quotes(self, section_11_subdivided):
        section = Enactment(**section_11_subdivided)
        quotes = [
            TextQuoteSelector(
                exact="The Department of Beards may issue licenses to such"
            ),
            TextQuoteSelector(exact="hairdressers", suffix=", or other male grooming"),
            TextQuoteSelector(exact="as they see fit"),
        ]
        positions = section.convert_quotes_to_position(quotes)
        assert positions == TextPositionSet(
            positions=[
                TextPositionSelector(start=0, end=51),
                TextPositionSelector(start=61, end=73),
                TextPositionSelector(start=112, end=127),
            ],
        )

    @pytest.mark.vcr
    def test_text_sequence_has_no_consecutive_Nones(self, test_client):

        amend_14 = test_client.read(query="/us/const/amendment/XIV")
        selector = TextQuoteSelector(exact="life, liberty, or property")
        passage = amend_14.select(selector)
        selected_list = passage.text_sequence()

        assert len(selected_list) == 3
        assert selected_list[0] is None
        assert selected_list[1].text == "life, liberty, or property"
        assert selected_list[2] is None

    def test_select_with_position_selector(self, section_11_together):
        section = Enactment(**section_11_together)
        passage = section.select(TextPositionSelector(start=29, end=43))
        assert passage.selected_text() == "…issue licenses…"

    def test_invalid_selector_text(self, section_11_subdivided):
        selector = TextQuoteSelector(exact="text that doesn't exist in the code")
        enactment = Enactment(**section_11_subdivided)

        with pytest.raises(TextSelectionError):
            enactment.select(selector)

    def test_select_text_with_string(self, fourth_a):
        section = Enactment(**fourth_a)
        passage = section.select("The right of the people")
        assert passage.selected_text() == "The right of the people…"

    def test_select_method_clears_previous_selection(self, test_client, section_8):
        old_version = test_client.read_from_json(section_8["children"][1])
        old_selector = TextPositionSet(
            positions=TextPositionSelector(start=0, end=65),
        )
        passage = old_version.select(old_selector)
        assert passage.selected_text() == (
            "Any such person issued a notice to remedy under subsection 1 must…"
        )

    @pytest.mark.vcr
    def test_no_space_before_ellipsis(self, test_client):
        enactment = test_client.read(query="/us/usc/t17/s102/b")
        passage = enactment.select(TextQuoteSelector(suffix="idea, procedure,"))
        assert " …" not in passage.selected_text()

    @pytest.mark.vcr
    def test_select_near_end_of_section(self):
        amendment = self.client.read(query="/us/const/amendment/XIV")
        selector = TextPositionSelector(start=1920, end=1980)
        passage = amendment.select(selector)
        assert "The validity of the public debt" in passage.selected_text()
        set_passage = amendment.select(
            TextPositionSet(
                positions=[
                    TextPositionSelector(start=1921, end=1973),
                    TextPositionSelector(start=2111, end=2135),
                ]
            )
        )
        expected = "…The validity of the public debt of the United States…shall not be questioned.…"
        assert set_passage.selected_text() == expected


class TestCompareEnactment:
    @pytest.mark.vcr
    def test_equal_enactment_text(self, test_client):
        """Test provisions with the same text (different dates)."""
        old_version = test_client.read(query="/test/acts/47/6A", date=date(1999, 1, 1))
        new_version = test_client.read(query="/test/acts/47/6A", date=date(2020, 1, 1))
        assert old_version.means(new_version)

    @pytest.mark.vcr
    def test_unequal_enactment_text(self, fourth_a):

        enactment = Enactment(**fourth_a)
        selector = TextQuoteSelector(suffix=", and no Warrants")
        search_clause = enactment.select(selector)

        whole_provision = enactment.select_all()

        assert whole_provision != search_clause
        assert not whole_provision.means(search_clause)
        assert whole_provision >= search_clause
        assert whole_provision > search_clause

    @pytest.mark.vcr
    def test_not_gt_if_equal(self, test_client):
        """
        Test that __gt__ "implies" operator does not return True for equal Enactments.

        The ISO 8601 format should be valid for the date, even though a datetime
        would have been accepted.
        """
        enactment = test_client.read(query="/test/acts/47/1", date="1999-01-01")
        passage = enactment.select_all()
        assert passage == passage
        assert not passage > passage
        assert passage >= passage

    @pytest.mark.vcr
    def test_not_gt_if_equal_with_selection(self, test_client):

        search_clause = test_client.read(query="/us/const/amendment/IV")
        search_clause.select(TextQuoteSelector(suffix=", and no Warrants"))

        assert search_clause == search_clause
        assert search_clause.means(search_clause)
        assert not search_clause > search_clause

    @pytest.mark.vcr
    def test_different_section_same_text(self, test_client, old_section_8, section_8):

        old_version = test_client.read_from_json(
            old_section_8["children"][1]["children"][1]
        )
        new_version = test_client.read_from_json(
            section_8["children"][1]["children"][4]
        )
        assert old_version.means(new_version)

    @pytest.mark.vcr
    def test_combined_section_implies_subdivided_section(
        self, section_11_together, section_11_subdivided
    ):
        combined = Enactment(**section_11_together)
        passage = combined.select_all()
        subdivided = Enactment(**section_11_subdivided)
        divided_passage = subdivided.select_all()
        assert passage >= divided_passage
        assert not passage > divided_passage
        assert passage.text_sequence() >= divided_passage.text_sequence()

    def test_passage_does_not_imply_text_sequence(
        self, section_11_together, section_11_subdivided
    ):
        combined = Enactment(**section_11_together)
        passage = combined.select_all()
        subdivided = Enactment(**section_11_subdivided)
        divided_passage = subdivided.select_all()
        with pytest.raises(TypeError):
            passage >= divided_passage.text_sequence()

    @pytest.mark.vcr
    def test_more_provisions_implies_fewer(self, test_client, section_8):
        fewer_provisions = test_client.read(
            query="/test/acts/47/8/2", date=date(1999, 1, 1)
        )
        more_provisions = test_client.read_from_json(section_8["children"][1])
        assert more_provisions >= fewer_provisions
        assert more_provisions > fewer_provisions
        assert not fewer_provisions >= more_provisions
        assert not fewer_provisions > more_provisions

    @pytest.mark.vcr
    def test_enactment_subset(self, section_11_together):
        combined = Enactment(**section_11_together)
        passage = combined.select_all()
        selector = TextQuoteSelector(
            exact="barbers, hairdressers, or other male grooming professionals"
        )
        limited = combined.select(selector)
        assert not passage.means(limited)
        assert combined > limited

    def test_cannot_compare_passage_with_text_version(self, section_8):
        enactment = Enactment(**section_8)
        passage = enactment.select_all()
        with pytest.raises(TypeError):
            passage.means(enactment.text_version)

    @pytest.mark.vcr
    def test_same_phrase_different_provisions_same_meaning(self, test_client):

        amend_5 = test_client.read(query="/us/const/amendment/V")
        amend_14 = test_client.read(query="/us/const/amendment/XIV/1")
        selector = TextQuoteSelector(
            exact="life, liberty, or property, without due process of law"
        )
        left = amend_5.select(selector)
        right = amend_14.select(selector)
        assert left.means(right)

    @pytest.mark.vcr
    def test_same_phrase_different_provisions_implication(self, test_client):
        amend_5 = test_client.read(query="/us/const/amendment/V")
        amend_14 = test_client.read(query="/us/const/amendment/XIV/1")
        selector = TextQuoteSelector(
            exact="life, liberty, or property, without due process of law"
        )
        left = amend_5.select(selector)
        right = amend_14.select(selector)
        assert left >= right

    @pytest.mark.vcr
    def test_same_phrase_selected_in_nested_provision_same_meaning(self, test_client):

        amend_5 = test_client.read(query="/us/const/amendment/V")
        amend_14 = test_client.read(query="/us/const/amendment/XIV")
        selector = TextQuoteSelector(
            exact="life, liberty, or property, without due process of law"
        )
        left = amend_5.select(selector)
        right = amend_14.select(selector)
        assert left.means(right)

    @pytest.mark.vcr
    def test_same_phrase_selected_in_nested_provision_implication(self, test_client):

        amend_5 = test_client.read(query="/us/const/amendment/V")
        amend_14 = test_client.read(query="/us/const/amendment/XIV")
        selector = TextQuoteSelector(
            exact="life, liberty, or property, without due process of law"
        )
        left = amend_5.select(selector)
        right = amend_14.select(selector)
        assert left >= right

    def test_fail_to_check_enactment_implies_textsequence(self, section_11_subdivided):
        subdivided = Enactment(**section_11_subdivided)
        text = subdivided.text_sequence()
        with pytest.raises(TypeError):
            _ = subdivided >= text

    def test_fail_to_check_if_enactment_means_textpassage(self, section_11_subdivided):
        subdivided = Enactment(**section_11_subdivided)
        subdivided.select_all()
        text = subdivided.text_sequence()
        with pytest.raises(TypeError):
            _ = subdivided.means(text.passages[0])

    def test_fail_to_check_if_textpassage_means_enactment(self, section_11_subdivided):
        subdivided = Enactment(**section_11_subdivided)
        subdivided.select_all()
        text = subdivided.text_sequence()
        with pytest.raises(TypeError):
            _ = text.passages[0].means(subdivided)


class TestAddEnactments:
    def test_add_subset_nested_enactment(self, section_8, test_client):
        """Test that adding an included Enactment returns the same Enactment."""
        greater = test_client.read_from_json(section_8["children"][1])
        lesser = test_client.read_from_json(section_8["children"][1]["children"][0])

        greater_passage = greater.select_all()
        lesser_passage = lesser.select_all()

        combined = greater_passage + lesser_passage

        assert combined.means(greater_passage)

        assert combined.means(greater_passage)

    @pytest.mark.vcr
    def test_add_superset_nested_enactment(self, section_8, test_client):
        """Test that adding an included Enactment returns the same Enactment."""
        greater = test_client.read_from_json(section_8["children"][1])
        lesser = test_client.read_from_json(section_8["children"][1]["children"][0])
        greater_passage = greater.select_all()
        lesser_passage = lesser.select_all()

        combined = lesser_passage + greater_passage

        assert combined.means(greater_passage)

    @pytest.mark.vcr
    def test_add_enactment_to_passage(self, section_8, test_client):
        """Test that adding an included Enactment returns the same Enactment."""
        greater = test_client.read_from_json(section_8["children"][1])
        passage = greater.select(
            "Any such person issued a notice to remedy under subsection 1 must"
        )
        lesser = test_client.read_from_json(section_8["children"][1]["children"][0])

        combined = passage + lesser

        selected_text = combined.selected_text()
        assert "Any such person" in selected_text
        assert "must…shave" in selected_text

    def test_add_shorter_plus_longer(self, fourth_a):
        fourth_a = Enactment(**fourth_a)
        selector = TextQuoteSelector(suffix=", and no Warrants")
        amendment = fourth_a.select_all()
        search_clause = fourth_a.select(selector)

        greater_plus_lesser = amendment + search_clause

        assert greater_plus_lesser.text == amendment.text
        assert greater_plus_lesser.means(amendment)

        lesser_plus_greater = search_clause + amendment
        assert lesser_plus_greater.text == amendment.text
        assert lesser_plus_greater.means(amendment)

    def test_add_overlapping_text_selection(self, fourth_a):
        enactment = Enactment(**fourth_a)
        passage = enactment.select(TextQuoteSelector(suffix=", and no Warrants"))
        new = enactment.make_selection(
            TextQuoteSelector(
                exact="shall not be violated, and no Warrants shall issue,"
            )
        )
        passage.select_more_text_at_current_node(new)

        expected = (
            "against unreasonable searches and seizures, "
            + "shall not be violated, "
            + "and no Warrants shall issue,"
        )
        assert expected in passage.selected_text()

    def test_non_overlapping_text_selection(self, fourth_a):
        enactment = Enactment(**fourth_a)
        left = enactment.select("The right of the people to be secure in their persons")
        right = enactment.select("shall not be violated")
        left.select_more_text_at_current_node(right.selection)
        assert left.selected_text() == (
            "The right of the people to be secure in their persons…"
            "shall not be violated…"
        )

    def test_limit_selected_text(self, fourth_a):
        enactment = Enactment(**fourth_a)
        passage = enactment.select(
            "The right of the people to be secure in their persons"
        )
        passage.select_more("shall not be violated")
        assert passage.selected_text() == (
            "The right of the people to be secure in their persons…"
            "shall not be violated…"
        )
        passage.limit_selection(start=40)
        assert passage.selected_text() == "…their persons…shall not be violated…"

    def test_change_selection_to_all(self, fourth_a):
        enactment = Enactment(**fourth_a)
        passage = enactment.select("right of the people")
        assert passage.selected_text() == ("…right of the people…")
        passage.select_all()
        assert passage.selected_text().startswith("The right of the people to")

    def test_select_unavailable_text(self, fourth_a):
        fourth = Enactment(**fourth_a)
        with pytest.raises(TextSelectionError):
            fourth.select("right to privacy")

    @pytest.mark.vcr
    def test_add_selection_from_changed_section(self, test_client):
        old_version = test_client.read("/test/acts/47/6D/1", date="1935-04-01")
        new_version = test_client.read("/test/acts/47/6D/1", date="2013-07-18")
        old = old_version.select("bona fide religious")
        new = new_version.select("reasons.")
        new.select_more_text_from_changed_version(old)
        assert new.selected_text() == "…bona fide religious…reasons."

    def test_add_selection_from_changed_node_and_subnode(
        self, old_section_8, section_8, test_client
    ):
        """Test that text that has changed subsections can still be added."""
        old_version = test_client.read_from_json(old_section_8)
        new_version = test_client.read_from_json(section_8)
        new_selection = new_version.select(
            "Any such person issued a notice to remedy under subsection 1 must"
        )
        old_selection = old_version.select("obtain a beardcoin")
        combined = new_selection + old_selection
        assert combined.selected_text().endswith("must…obtain a beardcoin…")
        # Test that original Enactments unchanged
        assert "obtain a beardcoin" not in new_selection.selected_text()

    def test_add_overlapping_enactments(self, fourth_a):
        enactment = Enactment(**fourth_a)
        search = enactment.select(TextQuoteSelector(suffix=", and no Warrants"))
        warrant = enactment.select(
            TextQuoteSelector(
                exact="shall not be violated, and no Warrants shall issue,"
            )
        )
        combined = search + warrant

        passage = (
            "against unreasonable searches and seizures, "
            + "shall not be violated, "
            + "and no Warrants shall issue,"
        )
        assert passage in combined.text
        # Test that original Enactments unchanged
        assert "and no Warrants" not in search.selected_text()

    def test_get_recursive_selection(self, section_8, test_client):
        version = test_client.read_from_json(section_8["children"][1])
        selector = TextPositionSet(
            positions=TextPositionSelector(start=0, end=65),
        )
        passage = version.select(selector)
        passage.select_more("obtain a beardcoin from the Department of Beards")

        ranges = passage.selection.ranges()
        assert ranges[0].start == 0
        assert ranges[0].end == 65

        assert ranges[1].start == 218
        assert ranges[1].end == 266

        as_quotes = passage.selection.as_quotes(version.text)
        assert as_quotes[1].exact == "obtain a beardcoin from the Department of Beards"

    def test_add_selection_from_child_node(self, section_8, test_client):
        parent_version = test_client.read_from_json(section_8["children"][1])
        parent_selector = TextPositionSet(
            positions=TextPositionSelector(start=0, end=65),
        )
        passage = parent_version.select(parent_selector)
        passage.select_more("obtain a beardcoin from the Department of Beards")
        assert passage.selected_text() == (
            "Any such person issued a notice to remedy under subsection 1 must…"
            "obtain a beardcoin from the Department of Beards…"
        )

        child_version = test_client.read_from_json(
            section_8["children"][1]["children"][3]
        )
        child_passage = child_version.select()

        combined = passage + child_passage

        assert combined.selected_text() == (
            "Any such person issued a notice to remedy under subsection 1 must…"
            "remove the beard with a laser, or "
            "obtain a beardcoin from the Department of Beards…"
        )

    def test_add_selection_from_parent_node(self, section_8, test_client):
        parent_version = test_client.read_from_json(section_8["children"][1])
        child_version = test_client.read_from_json(
            section_8["children"][1]["children"][3]
        )

        parent_passage = parent_version.select(
            "Any such person issued a notice to remedy under subsection 1 must"
        )
        child_passage = child_version.select("remove the beard with a laser")

        combined = child_passage + parent_passage
        assert combined.selected_text() == (
            "Any such person issued a notice to remedy under subsection 1 must…"
            "remove the beard with a laser…"
        )
        # original Enactments should be unchanged
        assert (
            parent_passage.selected_text()
            == "Any such person issued a notice to remedy under subsection 1 must…"
        )
        assert child_passage.selected_text() == "remove the beard with a laser…"

    def test_fail_to_add_repeated_text_from_changed_version(
        self, section_8, old_section_8, test_client
    ):
        """Fail to place selection because "Department of Beards" occurs twice in this scope."""
        new_version = test_client.read_from_json(section_8)
        new_passage = new_version.select(
            TextQuoteSelector(
                prefix="Department of Beards, ", exact="Australian Federal Police"
            )
        )

        old_version = test_client.read_from_json(old_section_8)
        old_passage = old_version.select(
            TextQuoteSelector(prefix="officer of the ", exact="Department of Beards")
        )

        with pytest.raises(TextSelectionError):
            new_passage + old_passage

    @pytest.mark.vcr
    def test_fail_to_add_non_parent_or_child_enactment(self, test_client):
        left = test_client.read("/test/acts/47/1")
        right = test_client.read("/test/acts/47/2")
        left_passage = left.select()
        right_passage = right.select()
        with pytest.raises(ValueError):
            _ = left_passage + right_passage

    def test_fail_to_add_text_not_in_this_version(
        self, old_section_8, section_8, test_client
    ):
        """Fail to add selection from new version because it isn't in the old version."""
        old_version = test_client.read_from_json(old_section_8)
        new_version = test_client.read_from_json(section_8)

        old_passage = old_version.select(
            "Any such person issued a notice to remedy under subsection 1 must"
        )
        new_passage = new_version.select("remove the beard with a laser")
        with pytest.raises(TextSelectionError):
            old_passage + new_passage

    def test_fail_to_add_node_not_in_this_version(
        self, old_section_8, section_8, test_client
    ):
        """Fail to add new selection because its node isn't in the old version."""
        old_version = test_client.read_from_json(old_section_8["children"][1])
        new_version = test_client.read_from_json(
            section_8["children"][1]["children"][3]
        )
        old_passage = old_version.select()
        new_passage = new_version.select()
        with pytest.raises(TextSelectionError):
            old_passage + new_passage

    @pytest.mark.xfail()
    def test_locate_anchor_by_remembering_prefix(
        self, old_section_8, section_8, test_client
    ):
        """
        Place position selector by remembering prefix from TextQuoteSelector.

        Xfails because the prefix information is lost when the selector
        becomes a position selector.
        """
        old_version = test_client.read_from_json(old_section_8)
        new_version = test_client.read_from_json(section_8)
        new_version.select(
            TextQuoteSelector(
                prefix="Department of Beards, ", exact="Australian Federal Police"
            )
        )

        old_version.select(
            TextQuoteSelector(prefix="officer of the ", exact="Department of Beards")
        )

        combined = new_version + old_version
        assert combined.text == "…Department of Beards…Australian Federal Police…"

    def test_error_for_using_wrong_type_to_select_text(self, section_8, test_client):
        new_version = test_client.read_from_json(section_8)
        with pytest.raises(TypeError):
            new_version.select(date(2000, 1, 1))

    def test_get_rangedict(self, old_section_8, test_client):
        old_version = test_client.read_from_json(old_section_8)
        result = old_version.rangedict()
        memo = result[3]
        assert memo.content.startswith("Where an officer")
        assert memo.start_date == date(1935, 4, 1)

    def test_passage_start_date_is_latest_amendment(self, section_8, test_client):
        new_version = test_client.read_from_json(section_8)
        new_passage = new_version.select("remove the beard with electrolysis")
        assert new_passage.start_date() == date(2013, 7, 18)
        assert new_passage.end_date() is None

    def test_repealed_passage_end_date_is_earliest_found(
        self, old_section_8, test_client
    ):
        old_version = test_client.read_from_json(old_section_8)
        old_passage = old_version.select(
            TextQuoteSelector(prefix="officer of the ", exact="Department of Beards")
        )
        old_passage.select_more("obtain a beardcoin from the Department of Beards")
        # adding an earlier end date
        old_passage.enactment.children[1].children[2].end_date = date(2001, 1, 1)
        old_passage.select_more("within 14 days of such notice")
        assert old_passage.start_date() == date(1935, 4, 1)
        assert old_passage.end_date() == date(2001, 1, 1)

    def test_unable_to_add_subsection_with_new_text(
        self, old_section_8, section_8, test_client
    ):
        old_version = test_client.read_from_json(old_section_8)
        new_version = test_client.read_from_json(section_8)
        old_passage = old_version.select(
            TextQuoteSelector(prefix="officer of the ", exact="Department of Beards")
        )
        new_passage = new_version.select("remove the beard with electrolysis")
        with pytest.raises(TextSelectionError):
            old_passage + new_passage

    def test_add_string_as_selector(self, section_11_subdivided):
        section = Enactment(**section_11_subdivided)
        passage = section.select("The Department of Beards may issue licenses to such")
        more = passage + "hairdressers"
        assert (
            more.selected_text()
            == "The Department of Beards may issue licenses to such…hairdressers…"
        )

    def test_clear_selection(self, section_8):
        section = Enactment(**section_8)
        passage = section.select_all()
        passage.clear_selection()
        assert passage.selected_text() == ""

    def test_select_from_passage(self, section_8):
        section = Enactment(**section_8)
        passage = section.select_all()
        passage.select("Australian Defence Force")
        assert passage.selected_text() == "…Australian Defence Force…"

    def test_cannot_select_text_with_citation(self, section_11_subdivided):
        section = Enactment(**section_11_subdivided)
        cite = section.as_citation()
        with pytest.raises(ValidationError):
            section.select(cite)


class TestConsolidateEnactments:
    """Test function for combining a list of Enactments."""

    def test_consolidate_enactments(self, fourth_a):
        enactment = Enactment(**fourth_a)
        search_selector = TextQuoteSelector(suffix=", and no Warrants")
        search_clause = enactment.select(search_selector)

        warrants_selector = TextQuoteSelector(prefix="shall not be violated,")
        warrants_clause = enactment.select(warrants_selector)

        fourth_amendment = enactment.select_all()

        consolidated = consolidate_enactments(
            [fourth_amendment, search_clause, warrants_clause]
        )
        assert len(consolidated) == 1
        assert consolidated[0].means(fourth_amendment)

    @pytest.mark.vcr()
    def test_consolidate_adjacent_passages(self, test_client):
        copyright_clause = test_client.read("/us/const/article/I/8/8")
        copyright_statute = test_client.read("/us/usc/t17/s102/b").select_all()

        selection = copyright_clause.select(None)
        securing_for_authors = selection + (
            "To promote the Progress of Science and "
            "useful Arts, by securing for limited Times to Authors"
        )
        and_inventors = selection + "and Inventors"
        right_to_writings = (
            selection + "the exclusive Right to their respective Writings"
        )
        to_combine = [
            copyright_statute,
            securing_for_authors,
            and_inventors,
            right_to_writings,
        ]
        combined = consolidate_enactments(to_combine)
        assert len(combined) == 2
        assert any(
            law.selected_text().startswith("To promote the Progress")
            and law.selected_text().endswith("their respective Writings…")
            for law in combined
        )

    def test_do_not_consolidate_from_different_sections(self, fifth_a, fourteenth_dp):

        due_process_5 = Enactment(**fifth_a)
        due_process_14 = Enactment(**fourteenth_dp)

        passage = "life, liberty, or property, without due process of law"
        quote_5 = due_process_5.select(passage)
        quote_14 = due_process_14.select(passage)

        combined = consolidate_enactments([quote_5, quote_14])
        assert len(combined) == 2



"""Citations to codified citations."""

from datetime import date
from enum import IntEnum
import json
from typing import Dict, List, Optional, Tuple, Union

from pydantic.class_validators import validator, root_validator
from pydantic import BaseModel, Field


class CodeLevel(IntEnum):
    CONSTITUTION = 1
    STATUTE = 2
    REGULATION = 3
    COURT_RULE = 4


# Path parts known to indicate the level of law they refer to.
KNOWN_CODES: Dict[str, Dict[str, Tuple[str, CodeLevel]]] = {
    "test": {"acts": ("Test Acts", CodeLevel.STATUTE)},
    "us": {
        "const": ("U.S. Const.", CodeLevel.CONSTITUTION),
        "usc": ("U.S. Code", CodeLevel.STATUTE),
        "cfr": ("CFR", CodeLevel.REGULATION),
    },
    "us-ca": {
        "const": ("Cal. Const.", CodeLevel.CONSTITUTION),
        "code": ("Cal. Codes", CodeLevel.STATUTE),
        "ccr": ("Cal. Code Regs.", CodeLevel.REGULATION),
        "roc": ("Cal. Rules of Court", CodeLevel.COURT_RULE),
    },
}


def identify_code(jurisdiction: str, code: str) -> Tuple[str, str]:
    """Find code name and type based on USLM citation parts."""
    try:
        sovereign = KNOWN_CODES[jurisdiction]
    except KeyError:
        raise KeyError(f'"{jurisdiction}" is not a known jurisdiction identifier')

    try:
        code_name, code_level = sovereign[code]
    except KeyError:
        raise KeyError(f'"{code}" is not a known code identifier')

    return code_name, code_level


class Citation(BaseModel):
    r"""
    A citation style for referring to an :class:`~legislice.enactments.Enactment` in written text.

    Intended for use with `Citation Style Language
    (CSL) <https://citeproc-js.readthedocs.io/en/latest/csl-json/markup.html>`_.
    """

    jurisdiction: str
    code: str
    code_level_name: Optional[str] = None
    volume: Optional[str] = None
    section: Optional[str] = None
    revision_date: Optional[date] = None
    type: str = Field("legislation", const=True)

    @root_validator(pre=True)
    def validate_code(cls, obj):
        if obj.get("code"):
            obj["code"], obj["code_level_name"] = identify_code(
                jurisdiction=obj["jurisdiction"], code=obj["code"]
            )
        return obj

    @validator("volume")
    def validate_volume(cls, value):
        if value:
            value = value.lstrip("t")
        return value

    @validator("section")
    def validate_section(cls, value):
        if value and not value.startswith("sec. "):
            value = "sec. " + value.lstrip("s")
        return value

    @staticmethod
    def csl_date_format(revision_date: date) -> Dict[str, List[List[Union[str, int]]]]:
        """Convert event date to Citation Style Language format."""
        return {
            "date-parts": [
                [
                    str(revision_date.year),
                    revision_date.month,
                    revision_date.day,
                ]
            ]
        }

    def __str__(self):
        name = f"{self.volume} {self.code} {self.section}"
        if self.revision_date:
            name += f" ({self.revision_date.year})"
        return name.replace("sec.", "§")

    def csl_dict(self) -> Dict[str, Union[str, int, List[List[Union[str, int]]]]]:
        """Return citation as a Citation Style Language object."""
        result = self.dict()
        result["container-title"] = result.pop("code", None)
        event_date = result.pop("revision_date", None)
        if event_date:
            result["event-date"] = self.csl_date_format(event_date)

        return result

    def csl_json(self) -> str:
        """Return citation as Citation Style Language JSON."""
        obj = self.csl_dict()
        return json.dumps(obj)



import datetime
import os
from typing import Dict

from anchorpoint import TextQuoteSelector
from dotenv import load_dotenv
import pytest

from legislice.download import Client

load_dotenv()

API_ROOT = "https://authorityspoke.com/api/v1"
TOKEN = os.getenv("LEGISLICE_API_TOKEN")


@pytest.fixture(scope="module")
def vcr_config():
    return {
        # Replace the Authorization request header with "DUMMY" in cassettes
        "filter_headers": [("authorization", "DUMMY")],
    }


@pytest.fixture(scope="class")
def test_client() -> Client:
    client = Client(api_token=TOKEN)
    client.coverage["/us/usc"] = {
        "latest_heading": "United States Code (USC)",
        "first_published": datetime.date(1926, 6, 30),
        "earliest_in_db": datetime.date(2013, 7, 18),
        "latest_in_db": datetime.date(2020, 8, 8),
    }
    client.coverage["/test/acts"] = {
        "latest_heading": "Test Acts",
        "first_published": datetime.date(1935, 4, 1),
        "earliest_in_db": datetime.date(1935, 4, 1),
        "latest_in_db": datetime.date(2013, 7, 18),
    }
    return client


@pytest.fixture(scope="class")
def section6d():
    return {
        "heading": "Waiver of beard tax in special circumstances",
        "content": "",
        "children": [
            {
                "heading": "",
                "content": "The Department of Beards shall waive the collection of beard tax upon issuance of beardcoin under Section 6C where the reason the maintainer wears a beard is due to bona fide religious, cultural, or medical reasons.",
                "children": [],
                "end_date": None,
                "node": "/test/acts/47/6D/1",
                "start_date": "2013-07-18",
                "url": "http://127.0.0.1:8000/api/v1/test/acts/47/6D/1@2018-03-11/",
            },
            {
                "heading": "",
                "content": "The determination of the Department of Beards as to what constitutes bona fide religious or cultural reasons shall be final and no right of appeal shall exist.",
                "children": [],
                "end_date": None,
                "node": "/test/acts/47/6D/2",
                "start_date": "1935-04-01",
                "url": "http://127.0.0.1:8000/api/v1/test/acts/47/6D/2@2018-03-11/",
            },
        ],
        "end_date": None,
        "node": "/test/acts/47/6D",
        "start_date": "1935-04-01",
        "url": "http://127.0.0.1:8000/api/v1/test/acts/47/6D@2018-03-11/",
        "parent": "http://127.0.0.1:8000/api/v1/test/acts/47@2018-03-11/",
    }


@pytest.fixture()
def section_8():
    return {
        "heading": "Notice to remedy",
        "start_date": "1935-04-01",
        "node": "/test/acts/47/8",
        "text_version": None,
        "url": "https://authorityspoke.com/api/v1/test/acts/47/8/",
        "end_date": None,
        "children": [
            {
                "heading": "",
                "start_date": "1935-04-01",
                "node": "/test/acts/47/8/1",
                "text_version": {
                    "id": 1142679,
                    "url": "https://authorityspoke.com/api/v1/textversions/1142679/",
                    "content": "Where an officer of the Department of Beards, Australian Federal Police, state or territorial police, or military police of the Australian Defence Force finds a person to be wearing a beard within the territory of the Commonwealth of Australia, and that person fails or is unable to produce a beardcoin as proof of holding an exemption under section 6, that officer shall in the first instance issue such person a notice to remedy.",
                },
                "url": "https://authorityspoke.com/api/v1/test/acts/47/8/1/",
                "end_date": None,
                "children": [],
                "citations": [
                    {
                        "target_uri": "/test/acts/47/6",
                        "target_url": "https://authorityspoke.com/api/v1/test/acts/47/6/",
                        "target_node": 1386965,
                        "reference_text": "section 6",
                    }
                ],
            },
            {
                "heading": "",
                "start_date": "1935-04-01",
                "node": "/test/acts/47/8/2",
                "text_version": {
                    "id": 1142683,
                    "url": "https://authorityspoke.com/api/v1/textversions/1142683/",
                    "content": "Any such person issued a notice to remedy under subsection 1 must either:",
                },
                "url": "https://authorityspoke.com/api/v1/test/acts/47/8/2/",
                "end_date": None,
                "children": [
                    {
                        "heading": "",
                        "start_date": "1935-04-01",
                        "node": "/test/acts/47/8/2/a",
                        "text_version": {
                            "id": 1142680,
                            "url": "https://authorityspoke.com/api/v1/textversions/1142680/",
                            "content": "shave in such a way that they are no longer in breach of section 5, or",
                        },
                        "url": "https://authorityspoke.com/api/v1/test/acts/47/8/2/a/",
                        "end_date": None,
                        "children": [],
                        "citations": [
                            {
                                "target_uri": "/test/acts/47/5",
                                "target_url": "https://authorityspoke.com/api/v1/test/acts/47/5/",
                                "target_node": 1386964,
                                "reference_text": "section 5",
                            }
                        ],
                    },
                    {
                        "heading": "",
                        "start_date": "2013-07-18",
                        "node": "/test/acts/47/8/2/b",
                        "text_version": {
                            "id": 1142702,
                            "url": "https://authorityspoke.com/api/v1/textversions/1142702/",
                            "content": "remove the beard with electrolysis, or",
                        },
                        "url": "https://authorityspoke.com/api/v1/test/acts/47/8/2/b/",
                        "end_date": None,
                        "children": [],
                        "citations": [],
                    },
                    {
                        "heading": "",
                        "start_date": "2013-07-18",
                        "node": "/test/acts/47/8/2/b-con",
                        "text_version": None,
                        "url": "https://authorityspoke.com/api/v1/test/acts/47/8/2/b-con/",
                        "end_date": None,
                        "children": [],
                        "citations": [],
                    },
                    {
                        "heading": "",
                        "start_date": "2013-07-18",
                        "node": "/test/acts/47/8/2/c",
                        "text_version": {
                            "id": 1142703,
                            "url": "https://authorityspoke.com/api/v1/textversions/1142703/",
                            "content": "remove the beard with a laser, or",
                        },
                        "url": "https://authorityspoke.com/api/v1/test/acts/47/8/2/c/",
                        "end_date": None,
                        "children": [],
                        "citations": [],
                    },
                    {
                        "heading": "",
                        "start_date": "2013-07-18",
                        "node": "/test/acts/47/8/2/d",
                        "text_version": {
                            "id": 1142681,
                            "url": "https://authorityspoke.com/api/v1/textversions/1142681/",
                            "content": "obtain a beardcoin from the Department of Beards",
                        },
                        "url": "https://authorityspoke.com/api/v1/test/acts/47/8/2/d/",
                        "end_date": None,
                        "children": [],
                        "citations": [],
                    },
                    {
                        "heading": "",
                        "start_date": "2013-07-18",
                        "node": "/test/acts/47/8/2/d-con",
                        "text_version": {
                            "id": 1142682,
                            "url": "https://authorityspoke.com/api/v1/textversions/1142682/",
                            "content": "within 14 days of such notice being issued to them.",
                        },
                        "url": "https://authorityspoke.com/api/v1/test/acts/47/8/2/d-con/",
                        "end_date": None,
                        "children": [],
                        "citations": [],
                    },
                ],
                "citations": [
                    {
                        "target_uri": "/test/acts/47/8/1",
                        "target_url": "https://authorityspoke.com/api/v1/test/acts/47/8/1/",
                        "target_node": 1386980,
                        "reference_text": "subsection 1",
                    }
                ],
            },
        ],
        "citations": [],
        "parent": "https://authorityspoke.com/api/v1/test/acts/47/",
    }


@pytest.fixture()
def old_section_8():
    return {
        "heading": "Notice to remedy",
        "start_date": "1935-04-01",
        "node": "/test/acts/47/8",
        "text_version": None,
        "url": "https://authorityspoke.com/api/v1/test/acts/47/8@1935-04-01",
        "end_date": None,
        "children": [
            {
                "heading": "",
                "start_date": "1935-04-01",
                "node": "/test/acts/47/8/1",
                "text_version": {
                    "id": 1142679,
                    "url": "https://authorityspoke.com/api/v1/textversions/1142679/",
                    "content": "Where an officer of the Department of Beards, Australian Federal Police, state or territorial police, or military police of the Australian Defence Force finds a person to be wearing a beard within the territory of the Commonwealth of Australia, and that person fails or is unable to produce a beardcoin as proof of holding an exemption under section 6, that officer shall in the first instance issue such person a notice to remedy.",
                },
                "url": "https://authorityspoke.com/api/v1/test/acts/47/8/1@1935-04-01",
                "end_date": None,
                "children": [],
                "citations": [
                    {
                        "target_uri": "/test/acts/47/6",
                        "target_url": "https://authorityspoke.com/api/v1/test/acts/47/6@1935-04-01",
                        "target_node": 1386965,
                        "reference_text": "section 6",
                    }
                ],
            },
            {
                "heading": "",
                "start_date": "1935-04-01",
                "node": "/test/acts/47/8/2",
                "text_version": {
                    "id": 1142683,
                    "url": "https://authorityspoke.com/api/v1/textversions/1142683/",
                    "content": "Any such person issued a notice to remedy under subsection 1 must either:",
                },
                "url": "https://authorityspoke.com/api/v1/test/acts/47/8/2@1935-04-01",
                "end_date": None,
                "children": [
                    {
                        "heading": "",
                        "start_date": "1935-04-01",
                        "node": "/test/acts/47/8/2/a",
                        "text_version": {
                            "id": 1142680,
                            "url": "https://authorityspoke.com/api/v1/textversions/1142680/",
                            "content": "shave in such a way that they are no longer in breach of section 5, or",
                        },
                        "url": "https://authorityspoke.com/api/v1/test/acts/47/8/2/a@1935-04-01",
                        "end_date": None,
                        "children": [],
                        "citations": [
                            {
                                "target_uri": "/test/acts/47/5",
                                "target_url": "https://authorityspoke.com/api/v1/test/acts/47/5@1935-04-01",
                                "target_node": 1386964,
                                "reference_text": "section 5",
                            }
                        ],
                    },
                    {
                        "heading": "",
                        "start_date": "1935-04-01",
                        "node": "/test/acts/47/8/2/b",
                        "text_version": {
                            "id": 1142681,
                            "url": "https://authorityspoke.com/api/v1/textversions/1142681/",
                            "content": "obtain a beardcoin from the Department of Beards",
                        },
                        "url": "https://authorityspoke.com/api/v1/test/acts/47/8/2/b@1935-04-01",
                        "end_date": "2013-07-18",
                        "children": [],
                        "citations": [],
                    },
                    {
                        "heading": "",
                        "start_date": "1935-04-01",
                        "node": "/test/acts/47/8/2/b-con",
                        "text_version": {
                            "id": 1142682,
                            "url": "https://authorityspoke.com/api/v1/textversions/1142682/",
                            "content": "within 14 days of such notice being issued to them.",
                        },
                        "url": "https://authorityspoke.com/api/v1/test/acts/47/8/2/b-con@1935-04-01",
                        "end_date": "2013-07-18",
                        "children": [],
                        "citations": [],
                    },
                ],
                "citations": [
                    {
                        "target_uri": "/test/acts/47/8/1",
                        "target_url": "https://authorityspoke.com/api/v1/test/acts/47/8/1@1935-04-01",
                        "target_node": 1386980,
                        "reference_text": "subsection 1",
                    }
                ],
            },
        ],
        "citations": [],
        "parent": "https://authorityspoke.com/api/v1/test/acts/47@1935-04-01",
    }


@pytest.fixture(scope="class")
def section_11_together():
    return {
        "heading": "Licensed repurchasers of beardcoin",
        "text_version": {
            "id": 1142710,
            "url": "https://authorityspoke.com/api/v1/textversions/1142710/",
            "content": "The Department of Beards may issue licenses to such barbers, hairdressers, or other male grooming professionals as they see fit to purchase a beardcoin from a customer whose beard they have removed, and to resell those beardcoins to the Department of Beards.",
        },
        "children": [],
        "end_date": "2013-07-18",
        "node": "/test/acts/47/11",
        "start_date": "1935-04-01",
        "url": "https://authorityspoke.com/api/v1/test/acts/47/11@1999-01-01",
        "parent": "https://authorityspoke.com/api/v1/test/acts/47@1999-01-01",
    }


@pytest.fixture(scope="class")
def citation_to_6c():
    return {
        "target_uri": "/test/acts/47/6C",
        "target_url": f"{API_ROOT}/test/acts/47/6C@1940-01-01",
        "target_node": 1660695,
        "reference_text": "Section 6C",
    }


@pytest.fixture(scope="function")
def section_11_subdivided():
    return {
        "heading": "Licensed repurchasers of beardcoin",
        "start_date": "2013-07-18",
        "node": "/test/acts/47/11",
        "text_version": {
            "id": 1142710,
            "url": "https://authorityspoke.com/api/v1/textversions/1142710/",
            "content": "The Department of Beards may issue licenses to such",
        },
        "url": "https://authorityspoke.com/api/v1/test/acts/47/11/",
        "end_date": None,
        "children": [
            {
                "heading": "",
                "start_date": "2013-07-18",
                "node": "/test/acts/47/11/i",
                "text_version": {
                    "id": 1142704,
                    "url": "https://authorityspoke.com/api/v1/textversions/1142704/",
                    "content": "barbers,",
                },
                "url": "https://authorityspoke.com/api/v1/test/acts/47/11/i/",
                "end_date": None,
                "children": [],
                "citations": [],
            },
            {
                "heading": "",
                "start_date": "2013-07-18",
                "node": "/test/acts/47/11/ii",
                "text_version": {
                    "id": 1142705,
                    "url": "https://authorityspoke.com/api/v1/textversions/1142705/",
                    "content": "hairdressers, or",
                },
                "url": "https://authorityspoke.com/api/v1/test/acts/47/11/ii/",
                "end_date": None,
                "children": [],
                "citations": [],
            },
            {
                "heading": "",
                "start_date": "2013-07-18",
                "node": "/test/acts/47/11/iii",
                "text_version": {
                    "id": 1142706,
                    "url": "https://authorityspoke.com/api/v1/textversions/1142706/",
                    "content": "other male grooming professionals",
                },
                "url": "https://authorityspoke.com/api/v1/test/acts/47/11/iii/",
                "end_date": None,
                "children": [],
                "citations": [],
            },
            {
                "heading": "",
                "start_date": "2013-07-18",
                "node": "/test/acts/47/11/iii-con",
                "text_version": {
                    "id": 1142707,
                    "url": "https://authorityspoke.com/api/v1/textversions/1142707/",
                    "content": "as they see fit to purchase a beardcoin from a customer",
                },
                "url": "https://authorityspoke.com/api/v1/test/acts/47/11/iii-con/",
                "end_date": None,
                "children": [],
                "citations": [],
            },
            {
                "heading": "",
                "start_date": "2013-07-18",
                "node": "/test/acts/47/11/iv",
                "text_version": {
                    "id": 1142708,
                    "url": "https://authorityspoke.com/api/v1/textversions/1142708/",
                    "content": "whose beard they have removed,",
                },
                "url": "https://authorityspoke.com/api/v1/test/acts/47/11/iv/",
                "end_date": None,
                "children": [],
                "citations": [],
            },
            {
                "heading": "",
                "start_date": "2013-07-18",
                "node": "/test/acts/47/11/iv-con",
                "text_version": {
                    "id": 1142709,
                    "url": "https://authorityspoke.com/api/v1/textversions/1142709/",
                    "content": "and to resell those beardcoins to the Department of Beards.",
                },
                "url": "https://authorityspoke.com/api/v1/test/acts/47/11/iv-con/",
                "end_date": None,
                "children": [],
                "citations": [],
            },
        ],
        "citations": [],
        "parent": "https://authorityspoke.com/api/v1/test/acts/47/",
    }


@pytest.fixture(scope="module")
def make_selector() -> Dict[str, TextQuoteSelector]:
    return {
        "bad_selector": TextQuoteSelector(exact="text that doesn't exist in the code"),
        "preexisting material": TextQuoteSelector(
            exact=(
                "protection for a work employing preexisting material in which "
                + "copyright subsists does not extend to any part of the work in "
                + "which such material has been used unlawfully."
            )
        ),
        "copyright": TextQuoteSelector(suffix="idea, procedure,"),
        "copyright_requires_originality": TextQuoteSelector(
            suffix="fixed in any tangible"
        ),
    }


@pytest.fixture(scope="class")
def first_a(test_client):
    return test_client.read_from_json(
        {
            "heading": "AMENDMENT I.",
            "start_date": "1791-12-15",
            "node": "/us/const/amendment/I",
            "text_version": {
                "id": 735703,
                "url": "https://authorityspoke.com/api/v1/textversions/735703/",
                "content": "Congress shall make no law respecting an establishment of religion, or prohibiting the free exercise thereof; or abridging the freedom of speech, or of the press; or the right of the people peaceably to assemble, and to petition the Government for a redress of grievances.",
            },
            "url": "https://authorityspoke.com/api/v1/us/const/amendment/I/",
            "end_date": None,
            "children": [],
            "citations": [],
            "parent": "https://authorityspoke.com/api/v1/us/const/amendment/",
        }
    )


@pytest.fixture(scope="class")
def second_a(test_client):
    return test_client.read_from_json(
        {
            "heading": "AMENDMENT II.",
            "start_date": "1791-12-15",
            "node": "/us/const/amendment/II",
            "text_version": {
                "id": 735704,
                "url": "https://authorityspoke.com/api/v1/textversions/735704/",
                "content": "A well regulated Militia being necessary to the security of a free State, the right of the people to keep and bear arms, shall not be infringed.",
            },
            "url": "https://authorityspoke.com/api/v1/us/const/amendment/II/",
            "end_date": None,
            "children": [],
            "citations": [],
            "parent": "https://authorityspoke.com/api/v1/us/const/amendment/",
        }
    )


@pytest.fixture(scope="class")
def third_a(test_client):
    return test_client.read_from_json(
        {
            "heading": "AMENDMENT III.",
            "start_date": "1791-12-15",
            "node": "/us/const/amendment/III",
            "text_version": {
                "id": 735705,
                "url": "https://authorityspoke.com/api/v1/textversions/735705/",
                "content": "No soldier shall, in time of peace be quartered in any house, without the consent of the Owner, nor in time of war, but in a manner to be prescribed by law.",
            },
            "url": "https://authorityspoke.com/api/v1/us/const/amendment/III/",
            "end_date": None,
            "children": [],
            "citations": [],
            "parent": "https://authorityspoke.com/api/v1/us/const/amendment/",
        }
    )


@pytest.fixture(scope="class")
def fourth_a():
    return {
        "heading": "AMENDMENT IV.",
        "start_date": "1791-12-15",
        "node": "/us/const/amendment/IV",
        "text_version": {
            "id": 735706,
            "url": "https://authorityspoke.com/api/v1/textversions/735706/",
            "content": "The right of the people to be secure in their persons, houses, papers, and effects, against unreasonable searches and seizures, shall not be violated, and no Warrants shall issue, but upon probable cause, supported by Oath or affirmation, and particularly describing the place to be searched, and the persons or things to be seized.",
        },
        "url": "https://authorityspoke.com/api/v1/us/const/amendment/IV/",
        "end_date": None,
        "children": [],
        "citations": [],
        "parent": "https://authorityspoke.com/api/v1/us/const/amendment/",
    }


@pytest.fixture(scope="module")
def fourth_a_no_text_version():
    return {
        "heading": "AMENDMENT IV.",
        "start_date": "1791-12-15",
        "node": "/us/const/amendment/IV",
        "content": "The right of the people to be secure in their persons, houses, papers, and effects, against unreasonable searches and seizures, shall not be violated, and no Warrants shall issue, but upon probable cause, supported by Oath or affirmation, and particularly describing the place to be searched, and the persons or things to be seized.",
        "url": "https://authorityspoke.com/api/v1/us/const/amendment/IV/",
        "end_date": None,
        "children": [],
        "citations": [],
        "parent": "https://authorityspoke.com/api/v1/us/const/amendment/",
    }


@pytest.fixture(scope="module")
def fifth_a():
    return {
        "heading": "AMENDMENT V.",
        "start_date": "1791-12-15",
        "node": "/us/const/amendment/V",
        "text_version": {
            "id": 735707,
            "url": "https://authorityspoke.com/api/v1/textversions/735707/",
            "content": "No person shall be held to answer for a capital, or otherwise infamous crime, unless on a presentment or indictment of a Grand Jury, except in cases arising in the land or naval forces, or in the Militia, when in actual service in time of War or public danger; nor shall any person be subject for the same offence to be twice put in jeopardy of life or limb; nor shall be compelled in any Criminal Case to be a witness against himself; nor be deprived of life, liberty, or property, without due process of law; nor shall private property be taken for public use, without just compensation.",
        },
        "url": "https://authorityspoke.com/api/v1/us/const/amendment/V/",
        "end_date": None,
        "children": [],
        "citations": [],
        "parent": "https://authorityspoke.com/api/v1/us/const/amendment/",
    }


@pytest.fixture(scope="module")
def fourteenth_dp():
    return {
        "heading": "Citizenship: security and equal protection of citizens.",
        "start_date": "1868-07-28",
        "node": "/us/const/amendment/XIV/1",
        "text_version": {
            "id": 735717,
            "url": "https://authorityspoke.com/api/v1/textversions/735717/",
            "content": "All persons born or naturalized in the United States, and subject to the jurisdiction thereof, are citizens of the United States and of the State wherein they reside. No State shall make or enforce any law which shall abridge the privileges or immunities of citizens of the United States; nor shall any State deprive any person of life, liberty, or property, without due process of law; nor deny to any person within its jurisdiction the equal protection of the laws.",
        },
        "url": "https://authorityspoke.com/api/v1/us/const/amendment/XIV/1/",
        "end_date": None,
        "children": [],
        "citations": [],
        "parent": "https://authorityspoke.com/api/v1/us/const/amendment/XIV/",
    }


@pytest.fixture(scope="module")
def provision_with_text_anchor():
    return {
        "passage": {
            "enactment": {
                "node": "/us/usc/t17/s102/a",
                "name": "copyright protection provision",
            },
            "selection": {
                "quotes": {
                    "exact": "Copyright protection subsists, in accordance with this title, in original works of authorship fixed in any tangible medium of expression, now known or later developed, from which they can be perceived, reproduced, or otherwise communicated, either directly or with the aid of a machine or device."
                }
            },
        },
        "anchors": {
            "quotes": "qualify for copyright protection. |17 U.S.C. § 102(a)|."
        },
    }


@pytest.fixture(scope="class")
def copyright_clause(test_client):
    return test_client.read_from_json(
        {
            "heading": "Patents and copyrights.",
            "start_date": "1788-09-13",
            "node": "/us/const/article/I/8/8",
            "text_version": {
                "id": 735650,
                "url": "https://authorityspoke.com/api/v1/textversions/735650/",
                "content": "To promote the Progress of Science and useful Arts, by securing for limited Times to Authors and Inventors the exclusive Right to their respective Writings and Discoveries;",
            },
            "url": "https://authorityspoke.com/api/v1/us/const/article/I/8/8/",
            "end_date": None,
            "children": [],
            "citations": [],
            "parent": "https://authorityspoke.com/api/v1/us/const/article/I/8/",
        }
    )


@pytest.fixture(scope="class")
def copyright_statute(test_client):
    return test_client.read_from_json(
        {
            "heading": "",
            "start_date": "2013-07-18",
            "node": "/us/usc/t17/s102/b",
            "text_version": {
                "id": 1030580,
                "url": "https://authorityspoke.com/api/v1/textversions/1030580/",
                "content": "In no case does copyright protection for an original work of authorship extend to any idea, procedure, process, system, method of operation, concept, principle, or discovery, regardless of the form in which it is described, explained, illustrated, or embodied in such work.",
            },
            "url": "https://authorityspoke.com/api/v1/us/usc/t17/s102/b/",
            "end_date": None,
            "children": [],
            "citations": [],
            "parent": "https://authorityspoke.com/api/v1/us/usc/t17/s102/",
        }
    )



import datetime
from legislice import download
from legislice.enactments import CitingProvisionLocation, CrossReference
import os

from anchorpoint import TextQuoteSelector
from dotenv import load_dotenv
import pytest

from legislice.download import (
    Client,
    LegislicePathError,
    LegisliceTokenError,
    enactment_needs_api_update,
)
from legislice.enactments import InboundReference


load_dotenv()

TOKEN = os.getenv("LEGISLICE_API_TOKEN")
API_ROOT = "https://authorityspoke.com/api/v1"


class TestDownloadJSON:
    client = Client(api_token=TOKEN, api_root=API_ROOT)

    @pytest.mark.vcr()
    def test_fetch_section(self, test_client):
        url = self.client.url_from_enactment_path("/test/acts/47/1")
        response = self.client._fetch_from_url(url=url)

        # Test that there was no redirect from the API
        assert not response.history

        section = response.json()
        assert section["start_date"] == "1935-04-01"
        assert section["end_date"] is None
        assert section["heading"] == "Short title"

    def test_download_from_wrong_domain_raises_error(self, test_client):
        url = self.client.url_from_enactment_path("/test/acts/47/1")
        wrong_url = url.replace("authorityspoke.com", "pythonforlaw.com")
        with pytest.raises(ValueError):
            self.client._fetch_from_url(url=wrong_url)

    @pytest.mark.vcr()
    def test_fetch_current_section_with_date(self, test_client):
        url = self.client.url_from_enactment_path(
            "/test/acts/47/6D", date=datetime.date(2020, 1, 1)
        )
        response = self.client._fetch_from_url(url=url)

        # Test that there was no redirect from the API
        assert not response.history

        waiver = response.json()
        assert waiver["url"].endswith("acts/47/6D@2020-01-01/")
        assert waiver["children"][0]["start_date"] == "2013-07-18"

    @pytest.mark.vcr()
    def test_wrong_api_token(self):
        bad_client = Client(api_token="wr0ngToken")
        with pytest.raises(LegisliceTokenError):
            bad_client.fetch(query="/test/acts/47/1")

    @pytest.mark.vcr()
    def test_no_api_token(self):
        bad_client = Client()
        with pytest.raises(LegisliceTokenError):
            bad_client.fetch(query="/test/acts/47/1")

    @pytest.mark.vcr()
    def test_extraneous_word_token_before_api_token(self):
        extraneous_word_token = "Token " + TOKEN
        client = Client(api_token=extraneous_word_token, api_root=API_ROOT)
        s102 = client.fetch(query="/test/acts/47/1")
        assert s102["start_date"] == "1935-04-01"
        assert s102["end_date"] is None
        assert s102["heading"] == "Short title"

    @pytest.mark.vcr()
    def test_fetch_past_section_with_date(self, test_client):
        waiver = test_client.fetch(
            query="/test/acts/47/6D", date=datetime.date(1940, 1, 1)
        )
        assert waiver["url"].endswith("acts/47/6D@1940-01-01/")
        assert waiver["children"][0]["start_date"] == "1935-04-01"

    @pytest.mark.vcr()
    def test_fetch_cross_reference_to_old_version(self, test_client):
        """Test that statute can be fetched with a post-enactment date it was in effect."""
        reference = CrossReference(
            target_uri="/test/acts/47/8",
            target_url="https://authorityspoke.com/api/v1/test/acts/47/8@1935-04-01/",
            reference_text="section 8",
        )
        enactment = test_client.fetch_cross_reference(
            query=reference, date=datetime.date(1950, 1, 1)
        )
        assert enactment["start_date"] == "1935-04-01"
        assert enactment["url"].endswith("@1950-01-01/")

    @pytest.mark.vcr()
    def test_omit_terminal_slash(self, test_client):
        statute = test_client.fetch(query="us/usc/t17/s102/b/")
        assert not statute["node"].endswith("/")

    @pytest.mark.vcr()
    def test_add_omitted_initial_slash(self, test_client):
        statute = test_client.fetch(query="us/usc/t17/s102/b/")
        assert statute["node"].startswith("/")


class TestDownloadAndLoad:
    def test_make_enactment_from_citation(self, test_client, fourth_a):
        """
        Test fields for loaded Enactment.

        known_revision_date should indicate whether the start_date is known to be
        the date that the provision was revised in the USC.
        """

        result = test_client.read_from_json(fourth_a)
        assert result.text.endswith("persons or things to be seized.")
        assert result.known_revision_date is True

    @pytest.mark.vcr()
    def test_make_enactment_from_selector_without_code(self, test_client):
        selection = TextQuoteSelector(suffix=", shall be vested")
        art_3 = test_client.read(query="/us/const/article/III/1")
        passage = art_3.select(selection)
        text = passage.selected_text()

        assert text.startswith("The judicial Power")
        assert text.endswith("the United States…")

    @pytest.mark.vcr()
    def test_bad_uri_for_enactment(self, test_client):
        with pytest.raises(LegislicePathError):
            _ = test_client.read(query="/us/const/article-III/1")

    @pytest.mark.vcr()
    def test_unavailable_path_within_partial_match(self, test_client):
        """
        Test when a key appears to be an ancestor of the desired path, but isn't.

        This jargon may be a leftover from testing the mocks.
        """
        client = test_client
        with pytest.raises(LegislicePathError):
            client.read(
                query="/us/const/amendment/XIV/2/b", date=datetime.date(2010, 12, 15)
            )

    @pytest.mark.vcr
    def test_date_is_too_early(self, test_client):
        client = test_client
        with pytest.raises(LegislicePathError):
            client.read(query="/us/usc/t17/s102/a", date=datetime.date(2010, 12, 15))

    @pytest.mark.vcr()
    def test_chapeau_and_subsections_from_uslm_code(self, test_client):
        """
        Test that the selected_text includes the text of subsections.

        known_revision_date is not available on the subsection as well as
        the section.
        """
        definition = test_client.read(query="/test/acts/47/4")
        sequence = definition.text_sequence()
        assert str(sequence.strip()).endswith("below the nose.")
        assert definition.known_revision_date is True
        assert definition.children[0].known_revision_date is False

    @pytest.mark.vcr()
    def test_unknown_revision_date(self, test_client):
        """
        Test notation that enactment went into effect before start of the available data range.

        This test may begin to fail if earlier statute versions are
        loaded to the API's database.
        """
        enactment = test_client.read(query="/us/usc/t17/s103")
        assert enactment.known_revision_date is False
        assert enactment.children[0].known_revision_date is False

    @pytest.mark.vcr()
    def test_download_from_cross_reference(self, test_client):
        ref = CrossReference(
            target_uri="/test/acts/47/6C",
            target_url=f"{API_ROOT}/test/acts/47/6C@2020-01-01/",
            target_node=1660695,
            reference_text="Section 6C",
        )
        cited = test_client.fetch(ref)
        assert cited["text_version"]["content"].startswith(
            "Where an exemption is granted"
        )


class TestReadJSON:
    @pytest.mark.vcr()
    def test_read_from_cross_reference(self, test_client):
        """Test reading old version of statute by passing date param."""
        ref = CrossReference(
            target_uri="/test/acts/47/6D",
            target_url=f"{API_ROOT}/test/acts/47/6D",
            reference_text="Section 6D",
        )
        cited = test_client.read(ref, date="1950-01-01")
        assert "bona fide religious or cultural reasons." in cited.text

    @pytest.mark.vcr()
    def test_read_enactment_without_version_url(self, test_client):
        data = {
            "start_date": "1935-04-01",
            "selection": [
                {
                    "start": 0,
                    "end": 250,
                }
            ],
            "text_version": {
                "content": (
                    "Where the Department provides an exemption from the prohibition "
                    "in section 5, except as defined in section 6D, the person to whom "
                    "such exemption is granted shall be liable to pay to the Department "
                    "of Beards such fee as may be levied under section 6B."
                ),
                "id": None,
                "url": None,
            },
            "heading": "Levy of beard tax",
            "anchors": [],
            "children": [],
            "node": "/test/acts/47/6A",
            "end_date": None,
        }
        result = test_client.read_from_json(data)
        assert result.content.startswith("Where")

    @pytest.mark.vcr
    def test_check_db_coverage_when_reading(self):
        """
        Test whether default client can check DB coverage.

        This test would not work if pre-2013 sections were added to the DB.
        This test would not work with the test client because it doesn't
        check for USC coverage.
        """

        serialized = {
            "heading": "No retroactive effect",
            "start_date": "2013-07-18",
            "node": "/us/usc/t17/s1332",
            "text_version": {
                "id": 1032460,
                "url": "https://authorityspoke.com/api/v1/textversions/1032460/",
                "content": "Protection under this chapter shall not be available for any design that has been made public under section 1310(b) before the effective date of this chapter.",
            },
            "url": "https://authorityspoke.com/api/v1/us/usc/t17/s1332/",
            "end_date": None,
            "children": [],
            "citations": [],
            "parent": "https://authorityspoke.com/api/v1/us/usc/t17/",
        }
        client = download.Client(api_token=TOKEN, api_root=API_ROOT)
        enactment = client.read_from_json(serialized)
        assert enactment.known_revision_date is False


class TestInboundCitations:
    client = Client(api_token=TOKEN, api_root=API_ROOT)

    @pytest.mark.vcr()
    def test_fetch_inbound_citations_to_node(self, test_client):
        infringement_statute = test_client.read(
            query="/us/usc/t17/s501",
        )
        inbound_refs = test_client.fetch_citations_to(infringement_statute)
        period_ref = inbound_refs[0]["locations"][0]
        assert period_ref.get("text_version", {}).get("content") is None

    @pytest.mark.vcr()
    def test_fetch_inbound_citations_in_multiple_locations(self, test_client):
        """
        Test InboundReference with multiple "locations".

        The string should be something like:
        InboundReference to /us/usc/t2/s1301, from (/us/usc/t2/s60c-5/a/2/A 2013-07-18) and 2 other locations

        But it's not clear which of the three locations will be chosen.
        """

        definitions = "/us/usc/t2/s1301"
        inbound_refs = test_client.citations_to(definitions)
        period_ref = inbound_refs[0]
        assert str(period_ref).endswith("and 2 other locations")

    @pytest.mark.vcr()
    def test_read_inbound_citations_to_node(self, test_client):
        infringement_statute = test_client.read(
            query="/us/usc/t17/s501",
        )
        inbound_refs = test_client.citations_to(infringement_statute)
        assert inbound_refs[0].content.startswith(
            "Any person who distributes a phonorecord"
        )
        assert inbound_refs[1].content.startswith(
            "The relevant provisions of paragraphs (2)"
        )
        period_ref = inbound_refs[0].locations[0]
        assert isinstance(period_ref, CitingProvisionLocation)
        assert period_ref.node == "/us/usc/t17/s109/b/4"
        assert period_ref.start_date.isoformat() == "2013-07-18"

    @pytest.mark.vcr()
    def test_download_inbound_citations_from_uri(self, test_client):
        inbound_refs = test_client.citations_to("/us/usc/t17/s501")
        assert inbound_refs[0].content.startswith(
            "Any person who distributes a phonorecord"
        )

    @pytest.mark.vcr()
    def test_download_enactment_from_inbound_citation(self, test_client):
        reference = InboundReference(
            content="Any person who distributes...",
            reference_text="section 501 of this title",
            target_uri="/us/usc/t17/s501",
            locations=[
                CitingProvisionLocation(
                    heading="",
                    node="/us/usc/t17/s109/b/4",
                    start_date=datetime.date(2013, 7, 18),
                )
            ],
        )
        cited = test_client.read(reference)
        assert cited.node == "/us/usc/t17/s109/b/4"
        assert cited.start_date == datetime.date(2013, 7, 18)
        assert repr(reference).startswith("InboundReference(content=")

    @pytest.mark.vcr()
    def test_download_enactment_from_citing_location(self, test_client):

        location = CitingProvisionLocation(
            heading="",
            node="/us/usc/t17/s109/b/4",
            start_date=datetime.date(2013, 7, 18),
        )
        enactment = test_client.read(location)
        assert enactment.content.startswith("Any person who distributes")

    @pytest.mark.vcr()
    def test_enactment_downloaded_from_citing_location_has_text(self):

        inbound_refs = self.client.citations_to("/us/usc/t17/s501")
        assert str(inbound_refs[0]).startswith("InboundReference to /us/usc/t17/s501")
        assert inbound_refs[0].content.startswith(
            "Any person who distributes a phonorecord"
        )
        citing_enactment = self.client.read(inbound_refs[0])
        assert citing_enactment.node == "/us/usc/t17/s109/b/4"
        assert citing_enactment.text.startswith(
            "Any person who distributes a phonorecord"
        )

    def test_node_field_needed_to_update_enactment(self):
        barbers_without_node = {
            "heading": "",
            "content": "barbers,",
            "children": [],
            "end_date": None,
            "start_date": "2013-07-18",
            "url": "https://authorityspoke.com/api/v1/test/acts/47/11/i@2020-01-01",
        }
        with pytest.raises(ValueError):
            _ = enactment_needs_api_update(barbers_without_node)



"""EnactmentGroup class."""

from __future__ import annotations

import textwrap
from typing import List, Optional, Sequence, Tuple, Union

from legislice.enactments import Enactment, EnactmentPassage, consolidate_enactments
from pydantic import BaseModel, validator


def sort_passages(passages: List[EnactmentPassage]) -> List[EnactmentPassage]:
    """
    Sort EnactmentPassages in group, in place.

    Sorts federal before state; constitutional before statute before regulation;
    and then alphabetically
    """
    passages.sort(key=lambda x: x.node)
    passages.sort(key=lambda x: x.level)
    passages.sort(key=lambda x: x.is_federal, reverse=True)
    return passages


class EnactmentGroup(BaseModel):
    """Group of Enactments with comparison methods."""

    passages: List[EnactmentPassage] = []

    @validator("passages", pre=True)
    def consolidate_passages(
        cls,
        obj: Union[
            EnactmentGroup,
            Enactment,
            EnactmentPassage,
            List[Union[Enactment, EnactmentPassage]],
        ],
    ) -> List[EnactmentPassage]:
        if isinstance(obj, EnactmentGroup):
            return obj.passages
        if not isinstance(obj, List):
            obj = [obj]
        consolidated: List[EnactmentPassage] = consolidate_enactments(obj)
        return consolidated

    @validator("passages")
    def sort_passages(cls, obj: List[EnactmentPassage]) -> List[EnactmentPassage]:
        return sort_passages(obj)

    def _at_index(self, key: int) -> EnactmentPassage:
        return self.passages[key]

    def __getitem__(
        self, key: Union[int, slice]
    ) -> Union[EnactmentPassage, EnactmentGroup]:
        if isinstance(key, slice):
            start, stop, step = key.indices(len(self))
            return self.__class__(
                passages=[self._at_index(i) for i in range(start, stop, step)]
            )
        return self._at_index(key)

    def __iter__(self):
        yield from self.passages

    def __len__(self):
        return len(self.passages)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(list(self.passages))})"

    def __str__(self):
        result = "the group of Enactments:"
        indent = "  "
        for factor in self.passages:
            result += f"\n{textwrap.indent(str(factor), prefix=indent)}"
        return result

    def _add_group(self, other: EnactmentGroup) -> EnactmentGroup:
        combined = self.passages[:] + other.passages[:]
        return self.__class__(passages=combined)

    def __add__(
        self, other: Union[EnactmentGroup, Sequence[Enactment], Enactment]
    ) -> EnactmentGroup:
        """Combine two EnactmentGroups, consolidating any duplicate Enactments."""
        if isinstance(other, self.__class__):
            return self._add_group(other)
        to_add = self.__class__(passages=other)
        return self._add_group(to_add)

    def __ge__(self, other: Union[Enactment, EnactmentPassage, EnactmentGroup]) -> bool:
        """Test whether ``self`` implies ``other`` and ``self`` != ``other``."""
        return bool(self.implies(other))

    def __gt__(self, other: Union[Enactment, EnactmentPassage, EnactmentGroup]) -> bool:
        """Test whether ``self`` implies ``other`` and ``self`` != ``other``."""
        return bool(self.implies(other))

    def _implies_enactment(self, other: Union[Enactment, EnactmentPassage]) -> bool:
        return any(self_enactment.implies(other) for self_enactment in self)

    def _implies(self, other: EnactmentGroup) -> bool:
        return all(self._implies_enactment(other_law) for other_law in other)

    def implies(
        self, other: Union[Enactment, EnactmentPassage, EnactmentGroup]
    ) -> bool:
        """Determine whether self includes all the text of another Enactment or EnactmentGroup."""
        if isinstance(other, (Enactment, EnactmentPassage)):
            return self._implies_enactment(other)
        return self._implies(other)



"""Enactments, their text, and cross-references between them."""

from __future__ import annotations

from collections import deque
from copy import deepcopy

from datetime import date
from typing import Any, Dict, Sequence, List, Optional, Tuple, Union

from anchorpoint import TextQuoteSelector, TextPositionSelector
from anchorpoint.textselectors import TextPositionSet, TextSelectionError
from anchorpoint.textselectors import TextPositionSetFactory
from anchorpoint.textsequences import TextSequence
from anchorpoint.utils.ranges import Range, RangeDict

from legislice.citations import Citation, identify_code, CodeLevel

from pydantic import BaseModel, validator, root_validator

RawSelector = Union[str, Dict[str, str]]
RawEnactment = Dict[str, Union[Any, str, List[RawSelector]]]


class CrossReference(BaseModel):
    """
    A legislative provision's citation to another provision.

    :param target_uri:
        the path to the target provision from the document root.

    :param target_url:
        the URL to fetch the target provision from an API.

    :param reference_text:
        The text in the citing provision that represents the cross-reference.
        Generally, this text identifies the target provision.

    :param target_node:
        an identifier for the target URI in the API.
    """

    target_uri: str
    target_url: str
    reference_text: str
    target_node: Optional[int] = None

    def __str__(self):
        return f'CrossReference(target_uri="{self.target_uri}", reference_text="{self.reference_text}")'


class CitingProvisionLocation(BaseModel):
    """
    Memo indicating where an Enactment can be downloaded.

    :param node:
        location of the citing provision in a legislative code

    :param start_date:
        start date of the citing version of the provision

    :param heading:
        heading text for the citing provision
    """

    node: str
    start_date: date
    heading: str = ""

    def __str__(self):
        return f"({self.node} {self.start_date})"

    def __lt__(self, other: CitingProvisionLocation):
        if self.start_date != other.start_date:
            return self.start_date < other.start_date
        return self.node < other.node


class InboundReference(BaseModel):
    """A memo that a TextVersion has a citation to a specified target provision."""

    content: str
    reference_text: str
    target_uri: str
    locations: List[CitingProvisionLocation]

    def __str__(self):
        result = f"InboundReference to {self.target_uri}, from {self.latest_location()}"
        if len(self.locations) > 1:
            result = f"{result} and {len(self.locations) - 1} other locations"
        return result

    def latest_location(self) -> CitingProvisionLocation:
        """Get most recent location where the citing text has been enacted."""
        return max(self.locations)

    @root_validator(pre=True)
    def search_citations_for_reference_text(
        cls, values: Dict[str, Union[Dict[str, str], str]]
    ) -> Dict:
        """Get reference_text field from nested "citations" model."""
        if not values.get("reference_text"):
            reference_text = ""
            for citation in values["citations"]:
                if citation["target_uri"] == values["target_uri"]:
                    reference_text = citation["reference_text"]
            values["reference_text"] = reference_text

        return values


class TextVersion(BaseModel):
    """Version of legislative text, enacted at one or more times and locations."""

    content: str
    url: Optional[str] = None
    id: Optional[int] = None

    @validator("content")
    def content_exists(cls, content: str) -> str:
        if not content:
            raise ValueError(
                "TextVersion should not be created with an empty string for content."
            )
        return content


class EnactmentMemo(BaseModel):

    """
    Info about an Enactment, to be linked to its text position in a parent Enactment.

    Used when recursively searching the parent Enactment.
    """

    node: str
    start_date: date
    content: str
    end_date: Optional[date] = None


class Enactment(BaseModel):
    """
    Base class for Enactments.

    Whether connected to subnodes by linking, or nesting.

    :param node:
        identifier for the site where the provision is codified

    :param heading:
        full heading of the provision

    :param text_version:
        full text content at this node, even if not all of it is cited

    :param start_date:
        date when the text was enacted at the cited location

    :param known_revision_date:
        whether the "start_date" field is known to be a date
        when the provision was revised in the code where it was publised.
        If False, then the Enactment refers to a time range when the text was
        codified, without asserting that the Enactment was not codified at
        earlier dates. This field is useful when working from incomplete
        legislative records.

    :param end_date:
        date when the text was removed from the cited location

    :param first_published:
        date when this Enactment's code was first published.

    :param earliest_in_db:
        date of the earliest version of this Enactment in the database.
        Used to determine whether the start_date of the Enactment is
        a date when the Enactment was amended or first published.

    :param anchors:
        a list of selectors representing the part of some other document
        (e.g. an Opinion) that references the Enactment. Unlike the selection
        field, it doesn't reference part of the Enactment itself. For use as
        a temporary place to store the anchors until they can be moved over
        to the citing document.

    :param name:
        a user-assigned label for the object
    """

    node: str
    start_date: date
    heading: str = ""
    text_version: Optional[TextVersion] = None
    end_date: Optional[date] = None
    first_published: Optional[date] = None
    earliest_in_db: Optional[date] = None
    anchors: Union[
        TextPositionSet, List[Union[TextPositionSelector, TextQuoteSelector]]
    ] = []
    citations: List[CrossReference] = []
    name: str = ""
    children: Union[List[Enactment], List[str]] = []

    @validator("text_version", pre=True)
    def make_text_version_from_str(
        cls, value: Optional[Union[TextVersion, str]]
    ) -> Optional[TextVersion]:
        """Allow content to be used to populate text_version."""
        if isinstance(value, str):
            if value:
                return TextVersion(content=value)
            return None
        return value or None

    @property
    def content(self) -> str:
        """Get text for this version of the Enactment."""
        if not self.text_version:
            return ""
        return self.text_version.content

    @property
    def nested_children(self):
        """Get nested children attribute."""
        return [child for child in self.children if isinstance(child, Enactment)]

    def get_identifier_part(self, index: int) -> Optional[str]:
        """Get a part of the split node identifier, by number."""
        identifier_parts = self.node.split("/")
        if len(identifier_parts) < (index + 1):
            return None
        return identifier_parts[index]

    @property
    def sovereign(self):
        """Get "sovereign" part of node identifier."""
        return self.get_identifier_part(1)

    @property
    def jurisdiction(self):
        """Get "sovereign" part of node identifier."""
        return self.sovereign

    @property
    def code(self):
        """Get "code" part of node identifier."""
        return self.get_identifier_part(2)

    @property
    def title(self):
        """Get "title" part of node identifier."""
        return self.get_identifier_part(3)

    @property
    def section(self):
        """Get "section" part of node identifier."""
        return self.get_identifier_part(4)

    @property
    def is_federal(self):
        """Check if self is from a federal jurisdiction."""
        return self.sovereign == "us"

    @property
    def level(self) -> str:
        """Get level of code for this Enactment, e.g. "statute" or "regulation"."""
        code_name, code_level_name = identify_code(self.sovereign, self.code)
        return code_level_name

    @property
    def padded_length(self):
        """Get length of self's content plus one character for space before next section."""
        if self.content:
            return len(self.content) + 1
        return 0

    @property
    def known_revision_date(self) -> bool:
        r"""
        Determine if Enactment's start_date reflects its last revision date.

        If not, then the `start_date` merely reflects the earliest date that versions
        of the :class:`Enactment`\'s code exist in the database.
        """
        if self.earliest_in_db:
            if self.earliest_in_db < self.start_date:
                return True
            elif self.first_published and self.earliest_in_db <= self.first_published:
                return True
        return False

    def __str__(self):
        return f"{self.node} ({self.start_date})"

    def as_citation(self) -> Citation:
        """Create Citation Style Language markup for the Enactment."""
        level = self.level
        if level != CodeLevel.STATUTE:
            raise NotImplementedError(
                f"Citation serialization not implemented for '{level}' provisions."
            )
        revision_date = self.start_date if self.known_revision_date else None
        return Citation(
            jurisdiction=self.jurisdiction,
            code=self.code,
            volume=self.title,
            section=self.section,
            revision_date=revision_date,
        )

    def cross_references(self) -> List[CrossReference]:
        """Return all cross-references from this node and subnodes."""
        result = self.citations[:]
        for child in self.nested_children:
            result += child.cross_references()
        return result

    def get_string(
        self,
        selection: Union[
            bool,
            str,
            TextPositionSelector,
            TextQuoteSelector,
            Sequence[TextQuoteSelector],
        ],
    ) -> str:
        """Use text selector to get corresponding string from Enactment."""
        position_set = self.convert_selection_to_set(selection)
        return position_set.as_string(text=self.text)

    def convert_selection_to_set(
        self,
        selection: Union[
            bool,
            str,
            TextPositionSelector,
            TextPositionSet,
            TextQuoteSelector,
            Sequence[TextQuoteSelector],
        ],
    ) -> TextPositionSet:
        """Create a TextPositionSet from a different selection method."""
        if selection is True:
            return self.make_selection_of_all_text()
        factory = TextPositionSetFactory(self.text)
        return factory.from_selection(selection)

    def convert_quotes_to_position(
        self, quotes: Sequence[TextQuoteSelector]
    ) -> TextPositionSet:
        """Convert quote selector to the corresponding position selector for this Enactment."""
        factory = TextPositionSetFactory(text=self.text)
        return factory.from_quote_selectors(quotes)

    def limit_selection(
        self,
        selection: TextPositionSet,
        start: int = 0,
        end: Optional[int] = None,
    ) -> TextPositionSet:
        """Limit selection to the range defined by start and end points."""

        limit_selector = TextPositionSelector.from_text(
            text=self.text, start=start, end=end
        )
        return selection & limit_selector

    def limit_selection_to_current_node(
        self, selection: TextPositionSet
    ) -> TextPositionSet:
        """Limit selection to the current node."""
        return self.limit_selection(selection=selection, start=0, end=len(self.text))

    def text_sequence(self, include_nones=True) -> TextSequence:
        """Get a sequence of text passages for this provision and its subnodes."""
        selection = self.tree_selection()
        return selection.as_text_sequence(text=self.text, include_nones=include_nones)

    def means(self, other: Union[Enactment, EnactmentPassage]) -> bool:
        """Determine if self and other have identical text."""
        if not isinstance(other, (EnactmentPassage, Enactment)):
            raise TypeError(
                f"Cannot compare {self.__class__.__name__} and {other.__class__.__name__} for same meaning."
            )
        self_selected_passages = self.text_sequence()
        other_selected_passages = other.text_sequence()
        return self_selected_passages.means(other_selected_passages)

    def select(
        self,
        selection: Union[
            bool,
            str,
            TextPositionSelector,
            TextPositionSet,
            TextQuoteSelector,
            Sequence[TextQuoteSelector],
        ] = True,
        start: int = 0,
        end: Optional[int] = None,
    ) -> EnactmentPassage:
        """Select text from Enactment."""
        selection_set = self.make_selection(selection=selection, start=start, end=end)
        return EnactmentPassage(enactment=self, selection=selection_set)

    def select_all(self) -> EnactmentPassage:
        """Return a passage for this Enactment, including all subnodes."""
        selection = self.make_selection_of_all_text()
        return EnactmentPassage(enactment=self, selection=selection)

    def make_selection_of_all_text(self) -> TextPositionSet:
        """Return a TextPositionSet of all text in this Enactment."""
        if self.text:
            return TextPositionSet(
                positions=TextPositionSelector(start=0, end=len(self.text))
            )
        return TextPositionSet()

    def make_selection(
        self,
        selection: Union[
            bool,
            str,
            TextPositionSelector,
            TextPositionSet,
            TextQuoteSelector,
            Sequence[TextQuoteSelector],
        ] = True,
        start: int = 0,
        end: Optional[int] = None,
    ) -> TextPositionSet:
        """
        Make a TextPositionSet for specified text in this Enactment.
        """
        if selection is False or selection is None:
            return TextPositionSet()
        elif not isinstance(selection, TextPositionSet):
            selection = self.convert_selection_to_set(selection)

        limited = self.limit_selection(selection=selection, start=start, end=end)
        self.raise_error_for_extra_selector(limited)
        return limited

    def raise_error_for_extra_selector(self, selection: TextPositionSet) -> None:
        """Raise an error if any passed selectors begin after the end of the text passage."""
        for selector in selection.positions:
            if selector.start > len(self.text) + 1:
                raise ValueError(f'Selector "{selector}" was not used.')

    def make_selection_of_this_node(self) -> TextPositionSet:
        """Return a TextPositionSet of the text at this node, not child nodes."""
        if not self.content:
            return TextPositionSet()
        return TextPositionSet(
            positions=TextPositionSelector(start=0, end=len(self.content))
        )

    def _rangedict(
        self, range_dict: RangeDict, tree_length: int
    ) -> Tuple[RangeDict, int]:
        if self.content:
            span = Range(start=tree_length, end=tree_length + len(self.content))
            range_dict[span] = EnactmentMemo(
                node=self.node,
                start_date=self.start_date,
                content=self.content,
                end_date=self.end_date,
            )
        new_tree_length = tree_length + self.padded_length

        for child in self.nested_children:
            range_dict, new_tree_length = child._rangedict(
                range_dict=range_dict, tree_length=new_tree_length
            )

        return range_dict, new_tree_length

    def rangedict(self) -> RangeDict:
        """Return a RangeDict matching text spans to Enactment attributes."""
        new_range_dict, new_tree_length = self._rangedict(
            range_dict=RangeDict(), tree_length=0
        )
        return new_range_dict

    @property
    def span_length(self) -> int:
        """Return the length of the span of this Enactment."""
        return self.padded_length + sum(
            child.span_length for child in self.nested_children
        )

    def _tree_selection(
        self, selector_set: TextPositionSet, tree_length: int
    ) -> Tuple[TextPositionSet, int]:
        selectors_at_node = self.make_selection_of_this_node()
        selectors_at_node_with_offset = selectors_at_node + tree_length
        new_tree_length = tree_length + self.padded_length
        new_selector_set = selector_set + selectors_at_node_with_offset

        for child in self.nested_children:
            new_selector_set, new_tree_length = child._tree_selection(
                selector_set=new_selector_set, tree_length=new_tree_length
            )

        return new_selector_set, new_tree_length

    def tree_selection(self) -> TextPositionSet:
        """Return set of selectors for selected text in this provision and its children."""
        new_selector_set, new_tree_length = self._tree_selection(
            selector_set=TextPositionSet(), tree_length=0
        )
        return new_selector_set

    def csl_json(self) -> str:
        """
        Serialize a citation to this provision in Citation Style Language JSON.

        Experimental feature.
        This CSL-JSON format currently only identifies the cited provision down to
        the section level. A citation to a subsection or deeper nested provision will
        be the same as a citation to its parent section.

        See https://citeproc-js.readthedocs.io/en/latest/csl-json/markup.html for a
        guide to this CSL-JSON format.
        """
        citation = self.as_citation()
        return citation.csl_json()

    @property
    def text(self):
        """Get all text including subnodes, regardless of which text is "selected"."""
        text_parts = [self.content]

        for child in self.nested_children:
            if child.text:
                text_parts.append(child.text)
        joined = " ".join(text_parts)
        return joined.strip()

    def implies(self, other: Union[Enactment, EnactmentPassage]) -> bool:
        """Test whether ``self`` has all the text passages of ``other``."""
        if not isinstance(other, (Enactment, EnactmentPassage)):
            raise TypeError(
                f"Cannot compare {self.__class__.__name__} and {other.__class__.__name__} for implication."
            )
        self_selected_passages = self.text_sequence(include_nones=False)
        other_selected_passages = other.text_sequence(include_nones=False)
        return self_selected_passages >= other_selected_passages

    def __ge__(self, other: Union[Enactment, EnactmentPassage]) -> bool:
        """
        Test whether ``self`` implies ``other``.

        :returns:
            Whether ``self`` contains at least all the same text as ``other``.
        """
        return self.implies(other)

    def __gt__(self, other) -> bool:
        """Test whether ``self`` implies ``other`` without having same meaning."""
        if self.means(other):
            return False
        return self.implies(other)


Enactment.update_forward_refs()


class EnactmentPassage(BaseModel):
    """An Enactment with selectors indicating which text is being referenced."""

    enactment: Enactment
    selection: TextPositionSet = TextPositionSet(
        positions=TextPositionSelector(start=0, end=None)
    )

    @property
    def text(self):
        """Get all text including subnodes, regardless of which text is "selected"."""
        return self.enactment.text

    @property
    def sovereign(self):
        """Get "sovereign" part of node identifier."""
        return self.enactment.sovereign

    @property
    def jurisdiction(self):
        """Get "sovereign" part of node identifier."""
        return self.enactment.sovereign

    @property
    def code(self):
        """Get "code" part of node identifier."""
        return self.enactment.code

    @property
    def title(self):
        """Get "title" part of node identifier."""
        return self.enactment.title

    @property
    def section(self):
        """Get "section" part of node identifier."""
        return self.enactment.section

    @property
    def is_federal(self):
        """Check if self is from a federal jurisdiction."""
        return self.enactment.is_federal

    @property
    def level(self) -> str:
        """Get level of code for this Enactment, e.g. "statute" or "regulation"."""
        return self.enactment.level

    @property
    def child_passages(self) -> List[EnactmentPassage]:
        """Return a list of EnactmentPassages for this Enactment's children."""
        result: List[EnactmentPassage] = []
        tree_length = self.enactment.padded_length
        for child in self.enactment.nested_children:
            selection = self.selection & TextPositionSelector(
                start=tree_length, end=tree_length + child.span_length
            )
            result.append(
                EnactmentPassage(enactment=child, selection=selection - tree_length)
            )
            tree_length += child.span_length
        return result

    def start_date(self):
        current = self.enactment.start_date
        ranges = self.enactment.rangedict()
        for span, memo in ranges.items():
            if self.selection.rangeset() & span[0]:
                if memo.start_date > current:
                    current = memo.start_date
        return current

    def end_date(self):
        current = self.enactment.end_date
        ranges = self.enactment.rangedict()
        for span, memo in ranges.items():
            if self.selection.rangeset() & span[0]:
                if not current:
                    current = memo.end_date
                elif memo.end_date and memo.end_date < current:
                    current = memo.end_date
        return current

    @property
    def node(self):
        """Get the node that this Enactment is from."""
        return self.enactment.node

    def as_quotes(self) -> List[TextQuoteSelector]:
        """Return quote selectors for the selected text."""
        return [
            phrase.as_quote(self.enactment.text) for phrase in self.selection.positions
        ]

    def __str__(self):
        text_sequence = self.text_sequence()
        return f'"{text_sequence}" ({self.enactment.node} {self.start_date()})'

    def select_all(self) -> None:
        """Select all text of Enactment."""
        text = self.enactment.text
        self.selection = TextPositionSet(
            positions=[TextPositionSelector(start=0, end=len(text))]
        )
        return None

    def limit_selection(
        self,
        start: int = 0,
        end: Optional[int] = None,
    ) -> None:
        """Limit selection to the range defined by start and end points."""

        limit_selector = TextPositionSelector.from_text(
            text=self.text, start=start, end=end
        )
        new_selection = self.selection & limit_selector
        self.selection = new_selection
        return None

    def select_more_text_at_current_node(
        self, added_selection: TextPositionSet
    ) -> None:
        """Select more text at this Enactment's node, not in child nodes."""
        new_selection = self.enactment.limit_selection_to_current_node(added_selection)
        self.selection = self.selection + new_selection
        return None

    def _update_text_at_included_node(
        self, other: EnactmentPassage
    ) -> Tuple[bool, bool]:
        """Recursively search child nodes for one that can be updated by `other`."""
        if self.node == other.node:
            found_node = True
            if self.text == other.text:
                self.select_more_text_in_current_branch(other.selection)
            else:
                self.select_more_text_from_changed_version(other)
            return found_node, True
        for selector in other.as_quotes():
            self.select_more(selector)
        return False, False

    def _add_passage_at_included_node(
        self, other: EnactmentPassage
    ) -> EnactmentPassage:
        """Add a selection of text at the same citation or at a child of self's citation."""

        if self >= other:
            return self

        copy_of_self = deepcopy(self)
        copy_of_self._update_text_at_included_node(other)
        return copy_of_self

    def implies(self, other: Union[Enactment, EnactmentPassage]) -> bool:
        """Test whether ``self`` has all the text passages of ``other``."""
        if not isinstance(other, (Enactment, EnactmentPassage)):
            raise TypeError(
                f"Cannot compare {self.__class__.__name__} and {other.__class__.__name__} for implication."
            )
        self_selected_passages = self.text_sequence(include_nones=False)
        other_selected_passages = other.text_sequence(include_nones=False)
        return self_selected_passages >= other_selected_passages

    def __add__(self, other: Union[Enactment, EnactmentPassage]) -> EnactmentPassage:

        if isinstance(other, Enactment):
            other = other.select_all()

        if not isinstance(other, self.__class__):
            copy_of_self = deepcopy(self)
            copy_of_self.select_more(other)
            return copy_of_self

        if other.node.startswith(self.node):
            return self._add_passage_at_included_node(other)
        elif self.node.startswith(other.node):
            return other._add_passage_at_included_node(self)

        raise ValueError(
            "Can't add selected text from two different Enactments "
            "when neither is a descendant of the other."
        )

    def __ge__(self, other: Union[Enactment, EnactmentPassage]) -> bool:
        """
        Test whether ``self`` implies ``other``.

        :returns:
            Whether ``self`` contains at least all the same text as ``other``.
        """
        return self.implies(other)

    def __gt__(self, other) -> bool:
        """Test whether ``self`` implies ``other`` without having same meaning."""
        if self.means(other):
            return False
        return self.implies(other)

    def select(
        self,
        selection: Union[
            bool,
            str,
            TextPositionSelector,
            TextPositionSet,
            TextQuoteSelector,
            Sequence[TextQuoteSelector],
        ] = True,
        start: int = 0,
        end: Optional[int] = None,
    ) -> None:
        """Select text from Enactment."""
        selection_set = self.enactment.make_selection(
            selection=selection, start=start, end=end
        )
        self.selection = selection_set
        return None

    def clear_selection(self) -> None:
        """Deselect any Enactment text, including in child nodes."""
        self.selection = TextPositionSet()

    def select_more_text_from_changed_version(self, other: EnactmentPassage) -> None:
        """
        Select more text from a different text version at the same citation path.

        :param other:
            An :class:`Enactment` representing different text enacted at a
            different time, at the same `node` (or USLM path citation) as self.
            This Element's :attr:`~Enactment.node` attribute must be the same
            string as self's node attribute. It's not sufficient for `other`
            to have an Enactment listed in its :attr:`~Enactment._children`
            attribute with the same node attribute,
            or for `other` to have the same node attribute as an ancestor of self.
        """
        incoming_quote_selectors = other.as_quotes()
        incoming_position_selectors = []
        for quote_selector in incoming_quote_selectors:
            position = quote_selector.as_unique_position(self.text)
            if position:
                incoming_position_selectors.append(position)
        self.select_more_text_in_current_branch(
            TextPositionSet(positions=incoming_position_selectors)
        )

    def selected_text(self) -> str:
        """
        Return this provision's text that is within the ranges described by self.selection.

        Based on creating an :class:`anchorpoint.textsequences.TextSequence` from this Enactment's
        text content and the ranges in its selection attribute.
        """
        text_sequence = self.text_sequence()
        return str(text_sequence)

    def text_sequence(self, include_nones: bool = True) -> TextSequence:
        """
        List the phrases in the Enactment selected by TextPositionSelectors.

        :param include_nones:
            Whether the list of phrases should include `None` to indicate a block of
            unselected text
        """
        selected = self.selection.as_text_sequence(
            text=self.enactment.text, include_nones=include_nones
        )
        return selected

    def select_more(
        self,
        selection: Union[
            str,
            TextPositionSelector,
            TextPositionSet,
            TextQuoteSelector,
            Sequence[TextQuoteSelector],
        ],
    ) -> None:
        """Select text, in addition to any previous selection."""
        if not isinstance(selection, TextPositionSet):
            selection = self.enactment.convert_selection_to_set(selection)

        self.selection += selection
        self.selection = self.selection.add_margin(text=self.text, margin_width=4)
        self.enactment.raise_error_for_extra_selector(selection)

    def select_more_text_in_current_branch(
        self, added_selection: TextPositionSet
    ) -> None:
        """Select more text within this Enactment's tree_selection, including child nodes."""
        new_selection = self.selection + added_selection
        self.selection = new_selection
        return None

    def means(self, other: EnactmentPassage) -> bool:
        r"""
        Find whether meaning of ``self`` is equivalent to that of ``other``.

        ``Self`` must be neither broader nor narrower than ``other`` to return True.

        :returns:
            whether ``self`` and ``other`` represent the same text
            issued by the same sovereign in the same level of
            :class:`Enactment`\.
        """
        if not isinstance(other, (EnactmentPassage, Enactment)):
            raise TypeError(
                f"Cannot compare {self.__class__.__name__} and {other.__class__.__name__} for same meaning."
            )
        self_selected_passages = self.text_sequence()
        other_selected_passages = other.text_sequence()
        return self_selected_passages.means(other_selected_passages)


class AnchoredEnactmentPassage(BaseModel):
    """A quoted Enactment passage with anchors to an external document.

    :param passage:
        an EnactmentPassage with selectors indicating which part
        of the Enactment is being referenced.

    :param anchor:
        anchors to text in an external document that references
        the EnactmentPassage
    """

    passage: EnactmentPassage
    anchors: TextPositionSet


def consolidate_enactments(
    enactments: Sequence[Union[Enactment, EnactmentPassage]]
) -> List[EnactmentPassage]:
    r"""
    Consolidate any overlapping :class:`Enactment`\s in a :class:`list`.

    :param enactments:
        a list of :class:`Enactment`\s that may include overlapping
        legislative text within the same section

    :returns:
        a list of :class:`Enactment`\s without overlapping text
    """
    consolidated: List[Enactment] = []
    enactments = [
        item.select_all() if isinstance(item, Enactment) else item
        for item in enactments
    ]
    while enactments:
        match_made = False
        left = enactments.pop()
        for right in enactments:
            try:
                combined = left + right
                enactments.remove(right)
                enactments.append(combined)
                match_made = True
                break
            except (ValueError, TypeError, TextSelectionError):
                pass
        if not match_made:
            consolidated.append(left)
    return consolidated



"""Download Enactments from API, with client."""

import datetime
from typing import Any, Dict, List, Mapping, Optional, Union

import requests

from legislice.enactments import (
    Enactment,
    CrossReference,
    CitingProvisionLocation,
    EnactmentPassage,
    InboundReference,
    RawEnactment,
)


def enactment_needs_api_update(data: RawEnactment) -> bool:
    """Determine if JSON representation of Enactment needs to be supplemented from API."""
    if not data.get("node"):
        raise ValueError(
            '"data" must contain a "node" field '
            "with a citation path to a legislative provision, "
            'for example "/us/const/amendment/IV"'
        )
    if data.get("heading") is None or data.get("start_date") is None:
        return True
    if data.get("content") is None and data.get("text_version") is None:
        return False
    return False


class LegislicePathError(Exception):
    """Error for API query with invalid path."""

    pass


class LegisliceDateError(Exception):
    """Error for API query with invalid date."""

    pass


class LegisliceTokenError(Exception):
    """Error for failure to authenticate with API token."""

    pass


def normalize_path(path: str) -> str:
    """Make sure path starts but does not end with a slash."""
    return "/" + path.strip("/")


class Client:
    """Downloader for legislative text."""

    def __init__(
        self,
        api_token: Optional[str] = "",
        api_root: Optional[str] = "https://authorityspoke.com/api/v1",
        update_coverage_from_api: bool = True,
    ):
        """Create download client with an API token and an API address."""
        self.api_root = api_root or ""

        if api_token and api_token.startswith("Token "):
            api_token = api_token.split("Token ")[1]
        self.api_token = api_token or ""
        self.coverage: Dict[str, Dict[str, Union[datetime.date, str]]] = {
            "/us/const": {
                "first_published": datetime.date(1788, 6, 21),
                "earliest_in_db": datetime.date(1788, 6, 21),
            }
        }
        self.update_coverage_from_api = update_coverage_from_api

    def fetch(
        self,
        query: Union[str, CitingProvisionLocation, CrossReference, InboundReference],
        date: Union[datetime.date, str] = "",
    ) -> RawEnactment:
        """
        Download legislative provision from string identifier or cross-reference.

        :param query:
            A cross-reference to the desired legislative provision, or a path to the
            desired legislation section using the United States Legislation Markup
            tree-like citation format.

        :param date:
            The date of the desired version of the provision to be downloaded. This is
            not needed if a :class:`~legislice.enactments.CrossReference` passed to the
            ``query`` param specifies a date. If no date is provided, the API will use the
            most recent date.
        """
        if isinstance(query, CrossReference):
            return self.fetch_cross_reference(query=query, date=date)
        elif isinstance(query, CitingProvisionLocation):
            return self.fetch_citing_provision(query=query)
        elif isinstance(query, InboundReference):
            return self.fetch_inbound_reference(query=query)
        return self.fetch_uri(query=query, date=date)

    def fetch_citing_provision(self, query: CitingProvisionLocation) -> RawEnactment:
        """
        Download legislative provision as Enactment from CitingProvisionLocation.

        CitingProvisionLocations are found in the `locations` attribute of the `InboundReference`
        objects obtained when using the `citations_to` method.
        """
        return self.fetch_uri(query=query.node, date=query.start_date)

    def fetch_cross_reference(
        self, query: CrossReference, date: Union[datetime.date, str] = ""
    ) -> RawEnactment:
        """
        Download legislative provision from cross-reference.

        :param query:
            A cross-reference to the desired legislative provision. Found by calling the
            :meth:`~legislice.enactments.Enactment.cross_references` method on an
            :class:`~legislice.enactments.Enactment` that contains one or more citations
            to other provisions.

        :param date:
            The date of the desired version of the provision to be downloaded. This is
            not needed if the :class:`~legislice.enactments.CrossReference` passed to the
            ``query`` param specifies a date. If no date is provided, the API will use the
            most recent date.
        """
        if isinstance(date, datetime.date):
            date = date.isoformat()

        target = query.target_url

        if date:
            if "@" in target:
                target = target.split("@")[0]
            target = f"{target}@{date}"

        return self._fetch_from_url(url=target).json()

    def fetch_db_coverage(self, code_uri: str) -> Dict[str, Union[str, datetime.date]]:
        """Document date range of provisions of a code of laws available in API database."""
        target = self.api_root + "/coverage" + code_uri
        coverage = self._fetch_from_url(url=target).json()
        for k, v in coverage.items():
            if k not in ("uri", "latest_heading"):
                coverage[k] = datetime.date.fromisoformat(v)
        return coverage

    def fetch_inbound_reference(self, query: InboundReference) -> RawEnactment:
        """
        Download legislative provision from InboundReference.

        :param query:
            An InboundReference identifying a provision containing a citation.

        :returns:
            An Enactment representing the provision containing the citation (not the cited provision).

        If the InboundReference has been enacted more than once, the latest one will be chosen.
        """
        most_recent = max(query.locations)
        return self.fetch_citing_provision(query=most_recent)

    def get_db_coverage(self, uri: str) -> str:
        """
        Add data about the API's coverage date range to the Enactment to be loaded.

        As a side effect, changes the Client's "coverage" attribute.

        :param uri:
            identifier for the Enactment to be created

        :returns:
            identifier for the Code of the Enactment to be created
        """
        uri_parts = uri.split("/")
        code_uri = ""
        if len(uri_parts) > 2:
            code_uri = f"/{uri_parts[1]}/{uri_parts[2]}"
            if self.update_coverage_from_api and not self.coverage.get(code_uri):
                self.coverage[code_uri] = self.fetch_db_coverage(code_uri)
        return code_uri

    def url_from_enactment_path(
        self, path: str, date: Union[datetime.date, str] = ""
    ) -> str:
        """Generate URL for API call for specified USLM path and date."""
        query_with_root = self.api_root + normalize_path(path)

        if isinstance(date, datetime.date):
            date = date.isoformat()
        if date:
            query_with_root = f"{query_with_root}@{date}"
        elif not query_with_root.endswith("/"):
            query_with_root += "/"
        return query_with_root

    def fetch_uri(
        self, query: str, date: Union[datetime.date, str] = ""
    ) -> RawEnactment:
        """
        Fetch data about legislation at specified path and date from Client's assigned API root.

        :param query:
            A path to the desired legislation section using the United States Legislation Markup
            tree-like citation format. Examples: "/us/const/amendment/IV", "/us/usc/t17/s103"

        :param date:
            A date when the desired version of the legislation was in effect. This does not need to
            be the "effective date" or the first date when the version was in effect. However, if
            you select a date when two versions of the provision were in effect at the same time,
            you will be given the version that became effective later.
        """
        url = self.url_from_enactment_path(path=query, date=date)
        response = self._fetch_from_url(url=url)
        return response.json()

    def uri_from_query(self, target: Union[str, Enactment, CrossReference]) -> str:
        """Get a URI for the target object."""
        if isinstance(target, Enactment):
            return target.node
        elif isinstance(target, CrossReference):
            return target.target_uri
        return target

    def fetch_citations_to(
        self, target: Union[str, Enactment, CrossReference]
    ) -> List[Dict]:
        """
        Query API for citations to a given target node, withoout loading them as InboundCitations.

        :param target:
            a string URI for the cited node, an Enactment at the cited node, or
            a CrossReference to the cited node

        :returns:
            a list of dicts representing citations to the cited node
        """
        uri = self.uri_from_query(target)
        query_with_root = self.api_root + "/citations_to" + uri
        api_response = self._fetch_from_url(query_with_root)
        return api_response.json()["results"]

    def citations_to(
        self, target: Union[str, Enactment, CrossReference]
    ) -> List[InboundReference]:
        r"""
        Load an InboundReference object for each provision citing the target USLM provision URI.

        :param target:
            a string URI for the cited node, an Enactment at the cited node, or
            a CrossReference to the cited node

        :returns:
            a list of InboundReferences to the cited node
        """
        target_uri = self.uri_from_query(target)
        json_citations = self.fetch_citations_to(target_uri)
        for entry in json_citations:
            entry["target_uri"] = target_uri
        return [InboundReference(**data) for data in json_citations]

    def update_data_from_api_if_needed(self, data: RawEnactment) -> RawEnactment:
        """
        Update data from API with data from API coverage.

        :param data:
            a dict representing the data from the API

        :returns:
            a dict representing the data from the API with updated data from API coverage
        """
        if enactment_needs_api_update(data):
            data = self.update_enactment_from_api(data)

        # update client's data about the database's coverage
        code_uri = self.get_db_coverage(data["node"])
        if self.coverage.get(code_uri):
            data["earliest_in_db"] = self.coverage[code_uri]["earliest_in_db"]
            data["first_published"] = self.coverage[code_uri]["first_published"]
        return data

    def read_passage_from_json(self, data: RawEnactment) -> EnactmentPassage:
        r"""
        Create a new :class:`EnactmentPassage` object using imported JSON data.

        If fields are missing from the JSON, they will be fetched using the API key.
        """

        data["enactment"] = self.update_data_from_api_if_needed(data=data["enactment"])
        if not data.get("selection"):
            data["selection"] = {"positions": [{"start": 0, "end": None}]}

        return EnactmentPassage(**data)

    def read_from_json(
        self, data: RawEnactment, use_text_expansion: bool = True
    ) -> Enactment:
        r"""
        Create a new :class:`Enactment` object using imported JSON data.

        If fields are missing from the JSON, they will be fetched using the API key.
        """

        data = self.update_data_from_api_if_needed(data=data)

        return Enactment(**data)

    def read(
        self,
        query: Union[str, CrossReference],
        date: Union[datetime.date, str] = "",
    ) -> Enactment:
        """
        Fetch data from Client's assigned API root and builds Enactment or LinkedEnactment.

        All text is selected by default.

        :param path:
            A path to the desired legislation section using the United States Legislation Markup
            tree-like citation format. Examples: "/us/const/amendment/IV", "/us/usc/t17/s103"

        :param date:
            A date when the desired version of the legislation was in effect. This does not need to
            be the "effective date" or the first date when the version was in effect. However, if
            you select a date when two versions of the provision were in effect at the same time,
            you will be given the version that became effective later.
        """
        raw_enactment = self.fetch(query=query, date=date)
        return self.read_from_json(raw_enactment, use_text_expansion=False)

    def update_enactment_from_api(self, data: RawEnactment) -> RawEnactment:
        """
        Use API to fill in missing fields in a dict representing an :class:`~legislice.enactments.Enactment`.

        Useful when the dict has missing data because it was created by a user.
        """

        data_from_api = self.fetch(query=data["node"], date=data.get("start_date"))
        return {**data, **data_from_api}

    def update_entries_in_enactment_index(
        self, enactment_index: Mapping[str, RawEnactment]
    ) -> Mapping[str, RawEnactment]:
        """Fill in missing fields in every entry in an :class:`~legislice.name_index.EnactmentIndex`."""
        for key, value in enactment_index.items():
            if enactment_needs_api_update(value["enactment"]):
                enactment_index[key]["enactment"] = self.update_enactment_from_api(
                    value["enactment"]
                )
        return enactment_index

    def _fetch_from_url(self, url: str) -> requests.Response:
        if not url.startswith(self.api_root):
            raise ValueError(
                f'target_url of cross-reference, "{url}", does not start with Client\'s api_root, "{self.api_root}"'
            )

        headers = {}
        if self.api_token:
            headers["Authorization"] = f"Token {self.api_token}"

        url = url.rstrip("/") + "/"

        response = requests.get(url, headers=headers)
        if response.status_code == 404:
            raise LegislicePathError(f"No enacted text found for query {url}")
        if response.status_code == 403:
            raise LegisliceTokenError(f"{response.json().get('detail')}")

        return response



from datetime import date

from anchorpoint import TextQuoteSelector
import pytest
from legislice.download import Client


class TestUpdateEnactments:
    """
    Tests for filling in blank fields in JSON Enactments with the test Client.

    The Enactments that need filling in could be user-generated as part of a model
    of an AuthoritySpoke holding, and can have some fields missing.
    The Enactments that come from the Client can be assumed not to be missing fields.
    """

    @pytest.mark.vcr
    def test_update_entries_in_enactment_index(self, test_client):
        enactment_index = {
            "security": {
                "enactment": {
                    "node": "/us/const/amendment/IV",
                    "start_date": "1791-12-15",
                },
                "selection": {
                    "quotes": [{"exact": "right of the people to be secure"}]
                },
            }
        }
        client = test_client
        updated_index = client.update_entries_in_enactment_index(enactment_index)
        updated_enactment = updated_index["security"]
        assert updated_enactment["enactment"]["heading"] == "AMENDMENT IV."
        assert updated_enactment["enactment"]["url"].startswith("https")

    @pytest.mark.vcr
    def test_update_entry_without_date(self, test_client):
        enactment_index = {
            "person clause": {
                "enactment": {"node": "/us/const/amendment/XIV/3"},
                "selection": {
                    "positions": [{"start": 3, "end": 15}],
                },
            }
        }
        client = test_client
        updated_index = client.update_entries_in_enactment_index(enactment_index)
        updated_passage = updated_index["person clause"]
        assert updated_passage["enactment"]["heading"].startswith(
            "Loyalty as a qualification"
        )
        loaded_enactment = client.read_passage_from_json(updated_passage)
        assert loaded_enactment.selected_text() == "…person shall…"

    @pytest.mark.vcr
    def test_read_enactment_with_suffix_field(self, test_client):
        raw_enactment = {
            "enactment": {
                "name": "search clause",
                "node": "/us/const/amendment/IV",
                "start_date": "1791-12-15",
            },
            "selection": {
                "quotes": {
                    "suffix": ", and no Warrants shall issue",
                }
            },
        }

        passage = test_client.read_passage_from_json(raw_enactment)
        assert passage.selected_text().endswith("shall not be violated…")

    @pytest.mark.vcr()
    def test_update_linked_enactment(self, test_client):
        data = {"node": "/us/const"}
        new = test_client.update_enactment_from_api(data)
        assert new["node"] == "/us/const"
        assert new["start_date"] == "1788-09-13"
        assert isinstance(new["children"][0], str)

    @pytest.mark.vcr()
    def test_update_enactment_when_reading_from_json(self, test_client):
        enactment = test_client.read_from_json(data={"node": "/us/const/amendment/IV"})
        assert enactment.start_date.isoformat() == "1791-12-15"

    @pytest.mark.vcr()
    def test_text_in_updated_enactment_is_selected_by_default(self, test_client):
        client = test_client
        passage = client.read_passage_from_json(
            data={"enactment": {"node": "/us/const/amendment/IV"}}
        )
        assert passage.selected_text().startswith("The right")


class TestLoadAndSelect:
    client = Client()
    client.coverage["/us/usc"] = {
        "earliest_in_db": date(1750, 1, 1),
        "first_published": date(1750, 1, 1),
    }
    response = {
        "heading": "",
        "start_date": "2013-07-18",
        "node": "/us/usc/t18/s1960/b/1",
        "text_version": {
            "id": 943740,
            "url": "https://authorityspoke.com/api/v1/textversions/943740/",
            "content": "the term “unlicensed money transmitting business” means a money transmitting business which affects interstate or foreign commerce in any manner or degree and—",
        },
        "url": "https://authorityspoke.com/api/v1/us/usc/t18/s1960/b/1/",
        "end_date": None,
        "children": [
            {
                "heading": "",
                "start_date": "2013-07-18",
                "node": "/us/usc/t18/s1960/b/1/A",
                "text_version": {
                    "id": 943737,
                    "url": "https://authorityspoke.com/api/v1/textversions/943737/",
                    "content": "is operated without an appropriate money transmitting license in a State where such operation is punishable as a misdemeanor or a felony under State law, whether or not the defendant knew that the operation was required to be licensed or that the operation was so punishable;",
                },
                "url": "https://authorityspoke.com/api/v1/us/usc/t18/s1960/b/1/A/",
                "end_date": None,
                "children": [],
                "citations": [],
            },
            {
                "heading": "",
                "start_date": "2013-07-18",
                "node": "/us/usc/t18/s1960/b/1/B",
                "text_version": {
                    "id": 943738,
                    "url": "https://authorityspoke.com/api/v1/textversions/943738/",
                    "content": "fails to comply with the money transmitting business registration requirements under section 5330 of title 31, United States Code, or regulations prescribed under such section; or",
                },
                "url": "https://authorityspoke.com/api/v1/us/usc/t18/s1960/b/1/B/",
                "end_date": None,
                "children": [],
                "citations": [],
            },
            {
                "heading": "",
                "start_date": "2013-07-18",
                "node": "/us/usc/t18/s1960/b/1/C",
                "text_version": {
                    "id": 943739,
                    "url": "https://authorityspoke.com/api/v1/textversions/943739/",
                    "content": "otherwise involves the transportation or transmission of funds that are known to the defendant to have been derived from a criminal offense or are intended to be used to promote or support unlawful activity;",
                },
                "url": "https://authorityspoke.com/api/v1/us/usc/t18/s1960/b/1/C/",
                "end_date": None,
                "children": [],
                "citations": [],
            },
        ],
        "citations": [],
        "parent": "https://authorityspoke.com/api/v1/us/usc/t18/s1960/b/",
    }

    def test_select_text_with_end_param(self):
        law = self.client.read_from_json(self.response)
        passage = law.select(
            TextQuoteSelector(suffix=", whether or not the defendant knew")
        )
        assert passage.selected_text().endswith("or a felony under State law…")

    def test_end_param_has_no_effect_when_nothing_selected(self):
        law = self.client.read_from_json(self.response)
        passage = law.select(selection=False, end="or a felony under State law")
        assert passage.selected_text() == ""



from copy import deepcopy
from datetime import date

from pydantic import ValidationError
import pytest

from legislice.enactments import Enactment
from legislice.groups import EnactmentGroup


class TestEnactmentGroups:
    def test_make_group(self, copyright_clause, copyright_statute):
        group = EnactmentGroup(passages=[copyright_clause, copyright_statute])
        assert len(group) == 2
        assert isinstance(EnactmentGroup(passages=group), EnactmentGroup)

    def test_consolidate_adjacent_passages(self, copyright_clause, copyright_statute):

        passage = copyright_clause.select(None)
        securing_for_authors = passage + (
            "To promote the Progress of Science and "
            "useful Arts, by securing for limited Times to Authors"
        )
        and_inventors = passage + "and Inventors"
        right_to_writings = passage + "the exclusive Right to their respective Writings"
        left = EnactmentGroup(passages=[and_inventors, copyright_statute])
        right = EnactmentGroup(passages=[securing_for_authors, right_to_writings])

        combined = left + right
        assert len(combined) == 2
        assert "/us/usc/t17/s102/b" in repr(combined)
        assert any(
            law.selected_text().startswith("To promote the Progress")
            and law.selected_text().endswith("their respective Writings…")
            for law in combined
        )

    def test_wrong_type_in_group(self, section6d, test_client):
        section = test_client.read_from_json(section6d)
        cite = section.as_citation()
        with pytest.raises(ValidationError):
            EnactmentGroup(passages=[cite])


class TestImplies:
    def test_no_implication_of_group(self, copyright_clause, copyright_statute):
        extra = deepcopy(copyright_clause)
        passage = extra.select(None)
        securing_for_authors = passage + (
            "To promote the Progress of Science and "
            "useful Arts, by securing for limited Times to Authors"
        )
        and_inventors = passage + "and Inventors"
        left = EnactmentGroup(passages=copyright_clause)
        right = EnactmentGroup(
            passages=[securing_for_authors, and_inventors, copyright_statute]
        )
        assert not left.implies(right)

    def test_implication_of_group(self, copyright_clause, copyright_statute):
        extra = deepcopy(copyright_clause)
        passage = extra.select(None)
        securing_for_authors = passage + (
            "To promote the Progress of Science and "
            "useful Arts, by securing for limited Times to Authors"
        )
        and_inventors = passage + "and Inventors"
        left = EnactmentGroup(passages=[copyright_clause, copyright_statute])
        right = EnactmentGroup(passages=[securing_for_authors, and_inventors])
        assert left.implies(right)
        assert left >= right

    def test_implication_of_enactment(self, copyright_clause, copyright_statute):
        extra = deepcopy(copyright_clause)
        passage = extra.select(None)
        securing_for_authors = passage + (
            "To promote the Progress of Science and "
            "useful Arts, by securing for limited Times to Authors"
        )
        left = EnactmentGroup(passages=[copyright_clause, copyright_statute])
        assert left > securing_for_authors


class TestAdd:
    def test_add_enactment_to_group(self, copyright_clause):
        passage = copyright_clause.select(None)
        securing_for_authors = passage + (
            "To promote the Progress of Science and "
            "useful Arts, by securing for limited Times to Authors"
        )
        and_inventors = passage + "and Inventors"
        right_to_writings = passage + "the exclusive Right to their respective Writings"
        left = EnactmentGroup(passages=[securing_for_authors, and_inventors])
        right = right_to_writings
        result = left + right
        assert len(result) == 1
        assert "respective Writings…" in str(result)
        assert "/us/const/article/I/8/8" in str(result[0])
        assert "1788-09-13" in str(result[:])

    def test_add_empty_group(self, copyright_clause, copyright_statute):
        passage = copyright_clause.select(None)
        securing_for_authors = passage + (
            "To promote the Progress of Science and "
            "useful Arts, by securing for limited Times to Authors"
        )
        writings = passage + "their respective Writings"
        left = EnactmentGroup(
            passages=[securing_for_authors, writings, copyright_statute]
        )
        assert len(left) == 2
        right = EnactmentGroup()
        result = left + right
        assert len(result) == 2

    def test_enactments_ordered_after_adding_groups(self, first_a, second_a, third_a):
        establishment_clause = first_a.select(
            "Congress shall make no law respecting an establishment of religion"
        )
        speech_clause = first_a.select(
            ["Congress shall make no law", "abridging the freedom of speech"]
        )

        arms_clause = second_a.select(
            "the right of the people to keep and bear arms, shall not be infringed."
        )

        left = EnactmentGroup(passages=[establishment_clause, arms_clause])
        right = EnactmentGroup(passages=[third_a, speech_clause])

        combined = left + right
        assert len(combined) == 3
        assert combined[0].node == "/us/const/amendment/I"

    def test_sort_enactments_in_group(self, copyright_clause, copyright_statute):
        regulation = Enactment(
            node="/us/cfr/t37/s202.1",
            heading="",
            start_date=date(1992, 2, 21),
            text_version="The following are examples of works not subject to copyright",
        )
        group = EnactmentGroup(
            passages=[regulation, copyright_clause, copyright_statute]
        )
        assert group[-1].node == "/us/cfr/t37/s202.1"

    def test_sort_state_enactment_in_group(self, copyright_clause, copyright_statute):
        regulation = Enactment(
            node="/us/cfr/t37/s202.1",
            heading="",
            start_date=date(1992, 2, 21),
            text_version="The following are examples of works not subject to copyright",
        )
        ca_statute = Enactment(
            node="/us-ca/code/evid/s351",
            start_date=date(1966, 1, 1),
            text_version="Except as otherwise provided by statute, all relevant evidence is admissible.",
            heading="",
        )
        group = EnactmentGroup(
            passages=[regulation, ca_statute, copyright_clause, copyright_statute]
        )
        assert group[-1].node == "/us-ca/code/evid/s351"



"""Utility for downloading and comparing the text of statutes and constitutional provisions."""

from anchorpoint.textselectors import TextQuoteSelector, TextPositionSelector
from anchorpoint.textselectors import TextPositionSet

from legislice.enactments import Enactment
from legislice.citations import Citation
from legislice.download import Client
from legislice.groups import EnactmentGroup

__version__ = "0.6.0"



import pytest

from legislice import Citation, Enactment


class TestMakeCitation:
    def test_bad_jurisdiction(self):
        with pytest.raises(KeyError):
            Citation(jurisdiction="atlantis", code="acts")

    def test_bad_code(self):
        with pytest.raises(KeyError):
            Citation(jurisdiction="us", code="proclamations")


class TestSerializeCitation:
    def test_make_citation_object(self, section6d, test_client):
        section = test_client.read_from_json(section6d)
        cite = section.as_citation()
        assert cite.revision_date.isoformat() == "1935-04-01"

    def test_constitutional_cite_not_implemented(self, fifth_a, test_client):
        amendment_5 = test_client.read_from_json(fifth_a)
        with pytest.raises(NotImplementedError):
            amendment_5.as_citation()

    def test_csl_format_citation_type(self, section6d, test_client):
        section = test_client.read_from_json(section6d)
        cite = section.as_citation()
        serialized = cite.csl_dict()
        assert serialized["type"] == "legislation"

    def test_csl_format_container_title(self, section6d, test_client):
        section = test_client.read_from_json(section6d)
        cite = section.as_citation()
        serialized = cite.csl_dict()
        assert serialized["container-title"] == "Test Acts"

    def test_csl_format_with_revision_date(self, section_11_subdivided, test_client):
        """Citation for provision with a revision date in the database."""
        section = test_client.read_from_json(section_11_subdivided)
        cite = section.as_citation()
        serialized = cite.csl_dict()
        assert str(cite) == "47 Test Acts § 11 (2013)"
        assert serialized.get("event-date") == {"date-parts": [["2013", 7, 18]]}
        assert serialized.get("revision_date") is None

    @pytest.mark.vcr
    def test_csl_format_no_revision_date(self, test_client):
        """
        Citation for provision with no known revision date in the database.

        The citation depth is limited to the section level.
        """
        section = test_client.read("/us/usc/t17/s103/b")
        cite = section.as_citation()
        serialized = cite.dict()
        assert str(cite).startswith("17 U.S. Code § 103")
        assert serialized.get("event-date") is None
        assert serialized.get("revision_date") is None

    @pytest.mark.vcr
    def test_usc_provision_with_revision_date(self, test_client):
        cares_act = test_client.read("/us/usc/t15/s9021/a/3/B/")
        cite = cares_act.as_citation()
        serialized = cite.csl_dict()
        assert str(cite) == "15 U.S. Code § 9021 (2020)"
        assert serialized["event-date"]["date-parts"][0] == ["2020", 4, 10]

    def test_citation_for_nested_provision(self, test_client, section_11_subdivided):
        """
        Both citations should be the same because of the CSL JSON format
        doesn't distinguish between citations below the section level.
        """
        section = test_client.read_from_json(section_11_subdivided)
        cite_json = section.as_citation().csl_dict()
        subsection_cite_json = section.children[0].as_citation().csl_dict()
        assert cite_json["section"] == subsection_cite_json["section"]
